# Author: Andrew Jewett (jewett.aij at g mail)
#         http://www.chem.ucsb.edu/~sheagroup
# License: 3-clause BSD License  (See LICENSE.TXT)
# Copyright (c) 2012, Regents of the University of California
# All rights reserved.


#__all__ = ['Ugraph', 'GraphMatcher', 'DFS', 'GenError', 'GraphError', 'Disconnected', 'NotUndirected']


import sys
import copy
from operator import itemgetter


class GenError(Exception):
    """
    An exception class containing string for error reporting.

    """

    def __init__(self, err_msg):
        self.err_msg = err_msg

    def __str__(self):
        return self.err_msg

    def __repr__(self):
        return str(self)


class GraphError(GenError):
    """
    An exception class containing a graph and a string for error reporting.

    """

    def __init__(self, g, err_msg):
        GenError.__init__(self, err_msg)
        self.g = g

    def __str__(self):
        g_str = str(g)
        # If the string representation of the graph is too
        # large to fit in one screen, truncate it
        g_str_lines = g_str.split('\n')
        if (len(g_str_lines) > 12):
            g_str_lines = g_str_lines[0:12] + \
                [' ...(additional lines not shown)]']
            g_str = '\n'.join(g_str_lines)
        return 'Problem with graph:\n' + g_str + '\n' + self.err_msg

    def __repr__(self):
        return str(self)


class Disconnected(GraphError):

    def __init__(self, g, err_msg):
        GraphError.__init__(self, g, err_msg)


class NotUndirected(GraphError):

    def __init__(self, g, err_msg):
        GraphError.__init__(self, g, err_msg)


class Edge(object):
    __slots__ = ["start", "stop", "attr"]

    def __init__(self,
                 iv_start,  # edge starts here (index into vertex list)
                 iv_stop,   # edge ends here (index into vertex list)
                 attr=None):  # edges have an optional type attribute
        self.start = iv_start
        self.stop = iv_stop
        self.attr = attr

    def __str__(self):
        return '(' + str(self.start) + ',' + str(self.stop) + ')'

    def __repr__(self):
        return str(self)


class Vertex(object):
    __slots__ = ["attr"]

    def __init__(self, attr=None):
        self.attr = attr


class Dgraph(object):
    """
    This class is a minimal implementation of a directed graph.
    Vertices and edges are accessed by integer index only (beginning at 0).
    Multiple edges connecting the same pair of vertices are allowed.
    (One would use the AddEdge() member function to accomplish this.)
    Both vertices and edges have an optional "attr" attribute.

    """

    NULL = -1  # forbidden vertex id number (used several places)

    def __init__(self, edgelist=None):
        """
        The constructor accepts an optional neighborlist argument.
        This is a simple list of neighbors for every vertex in the graph
        and it completely defines the topology of the graph.
        (Vertex and edge attributes can be specified later.)

        Alternatley, you can leave the neighborlist argument blank,
        and build the graph one vertex at a time later
        using the "AddVertex()" and "AddEdge()" commands.
        (AddEdge() commands must be issued strictly after
         all vertices have been defined.)

        """

        if edgelist == None:
            self.verts = []
            self.edges = []
            # integer keeps track of # of vertices = len(self.verts)
            self.nv = 0
            # integer keeps track of # of edges    = len(self.edges)
            self.ne = 0
            self.neighbors = []  # The adjacency list.

        else:
            # Parse the edge-list format:
            iv_max = 0  # <-- what's the vertex with the maximum id number?
            for i in range(0, len(edgelist)):
                iv = edgelist[i][0]
                jv = edgelist[i][1]
                if ((iv < 0) or (jv < 0)):
                    raise(GenError(
                        'Error in Dgraph.__init__: Negative vertex number pair encountered: (' + str(iv) + ',' + str(jv) + ')'))
                if iv > iv_max:
                    iv_max = iv
                if jv > iv_max:
                    iv_max = jv

            self.nv = iv_max + 1
            self.verts = [Vertex() for iv in range(0, self.nv)]
            self.edges = []
            self.ne = 0
            self.neighbors = [[] for iv in range(0, self.nv)]

            for i in range(0, len(edgelist)):
                iv = edgelist[i][0]
                jv = edgelist[i][1]
                self.neighbors[iv].append(self.ne)
                self.edges.append(Edge(iv, jv))
                self.ne += 1
            assert(self.ne == len(self.edges))

            self.SortNeighborLists()

    def AddVertex(self, iv=-1, attr=None):
        """
        Add a vertex to the graph.
        (Edges connected to this vertex must be added later using "AddEdge()"
         All vertices should be added before "AddEdge()" is ever invoked.)

        Optional "attr" argument allows you to set the attribute of this vertex.
        (for example, in a molecule this might correspond to the type of atom
         in the molecule).

        Optional "iv" argument allows you to specify the index of that vertex.
        Vertices can be added in any order, but thei vertex id numbers
        should eventually fill the range from 0 to self.nv-1.

        """
        if iv == -1:  # if iv unspecified, put the vertex at the end of the list
            iv = self.nv

        if iv < self.nv:
            self.verts[iv].attr = attr
        else:
            # In case there is a gap between iv and nv, fill it with blanks
            self.verts += ([Vertex()] * ((1 + iv) - self.nv))
            self.neighbors += ([[]] * ((1 + iv) - self.nv))
            self.verts[iv].attr = attr
            self.nv = iv + 1
            assert(self.nv == len(self.verts))
            assert(self.nv == len(self.neighbors))

    def AddEdge(self, iv, jv, attr=None, remove_duplicates=False):
        """
        Add an edge to graph connecting vertex iv to jv.
        (both are integers from 0 to self.nv-1)
        This function must not be called until all vertices have been added.
        If the edge is already present (and remove_duplicates==True),
        no new edge will be added.

        """
        if remove_duplicates:
            for je in self.neighbors[iv]:
                if jv == self.edges[je].stop:
                    return  # In that case, do nothing, the edge is already present
        self.edges.append(Edge(iv, jv, attr))
        self.neighbors[iv].append(self.ne)
        self.ne += 1
        assert(self.ne == len(self.edges))

    def ReorderVerts(self, vpermutation, invert=False):
        """
        This function allows the user to re-order (relabel) the vertices
        in a graph, making the necessary changes to the
        self.verts, self.edges, and self.neighbors lists.
        By default (invert=False).  The vpermutation is a list
        from 1 to self.nv which is interpreted this way:
            iv = vpermutation[iv_orig]
        where "iv" and "iv_orig" are the vertex id numbers before
        and after the mapping (which also corresponds to its
        position in the self.verts and self.neighbors arrays).

        """
        assert(len(self.verts) == self.nv)
        assert(len(self.edges) == self.ne)
        assert(len(vpermutation) == self.nv)

        if (invert):
            vperm = [-1 for iv in vpermutation]
            for iv in range(0, self.nv):
                vperm[vpermutation[iv]] = iv
        else:
            vperm = vpermutation

        orig_verts = [vert for vert in self.verts]
        for iv_old in range(0, self.nv):
            iv = vperm[iv_old]
            self.verts[iv] = orig_verts[iv_old]

        for ie in range(0, self.ne):
            self.edges[ie].start = vperm[self.edges[ie].start]
            self.edges[ie].stop = vperm[self.edges[ie].stop]

        orig_neighbors = [nlist for nlist in self.neighbors]
        # self.neighbors is a 2-d array.
        # We need to re-sort "self.neighbors" because the first index is
        # a vertex id number, and these id numbers have been permuted.
        # However, there's no need to sort the contents of each sub-array
        # (self.neighbors[iv]), because these are edge id numbers (indices into
        # the self.edges[] array). These edge index numbers are never altered.
        # (However the entries stored in self.edges were modified earlier.)
        for iv_old in range(0, self.nv):
            iv = vperm[iv_old]
            self.neighbors[iv] = orig_neighbors[iv_old]

        # Optional:
        self.SortNeighborLists()

    def ReorderEdges(self, epermutation, invert=False):
        """
        This function allows the user to re-order (relabel) the
        edges in a graph, making the necessary changes to the
        self.edges and self.neighbors lists.
        By default (invert=False).  The epermutation is a list
        from 1 to self.ne which is interpreted this way:
            ie = epermutation[ie_orig]
        where "ie" and "ie_orig" are the edge id numbers before
        and after the mapping (which also corresponds to that edge's
        position in the self.edges array).
            (Minor detail: Recall that in this code, Ugraphs
        are implemented by placing two (directed) edges between each pair of
        connected, adjacent vertices, which point back-and-forth between them.
        Consequently the list of edges in self.edges is often typically
        twice as large you might expect.)

        """
        assert(len(self.verts) == self.nv)
        assert(len(self.edges) == self.ne)
        assert(len(epermutation) == self.ne)

        if (invert):
            eperm = [-1 for ie in epermutation]
            for ie in range(0, self.ne):
                eperm[epermutation[ie]] = ie
        else:
            eperm = epermutation

        orig_edges = [edge for edge in self.edges]
        for ie_old in range(0, self.ne):
            ie = eperm[ie_old]
            self.edges[ie] = orig_edges[ie_old]

        for iv in range(0, self.nv):
            for j in range(0, len(self.neighbors[iv])):
                je_old = self.neighbors[iv][j]
                self.neighbors[iv][j] = eperm[je_old]

    def SortNeighborLists(self):
        assert(self.nv == len(self.neighbors))
        for iv in range(0, self.nv):
            # Back when self.neighbors was just a 2-dimensional list of
            # vertex id numbers, then the following line would have worked:
            #  self.neighbors[iv].sort()

            # ugly python code alert:
            # Unfortunately, we had to change the format of self.neighbors. Now
            # it is a list of indices into the self.edges array ("ie" numbers).
            # We want to sort the "ie" numbers by the vertices they point to.
            # self.edge[ie].start should point to the current vertex (hopefully).
            # self.edge[ie].stop should point to the vertex it's attached to.
            # So we want to sort the ie's in self.neighbors by self.edge[ie].stop
            # Create a temporary array of 2-tuples (ie, jv)
            nlist = [(ie, self.edges[ie].stop)
                     for ie in self.neighbors[iv]]
            self.neighbors[iv] = [ie for ie, jv in sorted(nlist,
                                                          key=itemgetter(1))]

    def FindEdge(self, istart, istop):
        """
        A simple function looks up the edge id number
        corresponding to an edge connecting vertex istart to istop.
        If not present returns Dgraph.NULL.

        """
        iv = istart
        for je in self.neighbors[iv]:
            jv = self.edges[je].stop
            if jv == istop:
                return je
        return Dgraph.NULL

    def GetVert(self, iv):
        return self.verts[iv]

    def GetEdge(self, ie):
        return self.edges[ie]

    def GetNumVerts(self):
        return self.nv

    def GetNumEdges(self):
        return self.ne

    # Commenting out.  I think it's clearer to use python's deepcopy instead
    # def makecopy(self):
    #    new_copy = Ugraph()
    #    new_copy.verts = [vertex for vertex in self.verts]
    #    new_copy.edges = [  edge for edge   in self.edges]
    #    new_copy.neighbors = [nlist for nlist in self.neighbors]
    #    new_copy.nv = self.nv
    #    new_copy.ne = self.ne
    #    return new_copy

    def __str__(self):
        # Print the graph as a list of neighbor-lists.
        # (Note: This is the same format as the first argument to __init__().
        #        The Vertex.attr and Edge.attr attributes are not printed.)
        l = ['([']
        for iv in range(0, self.nv):
            l.append('[')
            for j in range(0, len(self.neighbors[iv])):
                je = self.neighbors[iv][j]
                jv = self.edges[je].stop
                l.append(str(jv))
                if j < len(self.neighbors[iv]) - 1:
                    l.append(', ')
                else:
                    l.append(']')
            if iv < self.nv - 1:
                l.append(',\n  ')
            else:
                l.append(']')
        l.append(',\n [')
        for ie in range(0, self.ne):
            l.append(str(self.edges[ie]))
            if ie < self.ne - 1:
                l.append(', ')
            else:
                l.append('])\n')
        return ''.join(l)

    def __repr__(self):
        return str(self)


class Ugraph(Dgraph):
    """
    This class is a minimal implementation of an undirected graph.
    Vertices and edges are accessed by integer index only (beginning at 0).
    Multiple edges connecting the same pair of vertices are allowed.
    (One would use the AddEdge() member function to accomplish this.)
    Both vertices and edges have an optional "attr" attribute.

        Undirected graphs (Ugraphs) are represented internally as
        directed graphs.  This means that for every edge in the Ugraph,
        connecting vertex 2 to 3, for example, two edges are stored
        internally, (2 -> 3,   and   3 -> 2),
        Edges which begin and end at the same vertex are stored only once.)

    """

    def __init__(self, edgelist=None):
        Dgraph.__init__(self, edgelist)

        # Now add the extra edges which point in the reverse direction.
        neu = self.ne
        ned = self.ne
        for ieu in range(0, self.ne):
            iv = self.edges[ieu].start
            jv = self.edges[ieu].stop
            if iv != jv:
                ned += 1

        self.ieu_to_ied = [Dgraph.NULL for ieu in range(0, neu)]
        self.ied_to_ieu = [Dgraph.NULL for ied in range(0, ned)]

        ied_redundant = neu
        for ie in range(0, neu):
            iv = self.edges[ie].start
            jv = self.edges[ie].stop
            attr = self.edges[ie].attr
            self.ieu_to_ied[ie] = ie
            self.ied_to_ieu[ie] = ie

            if iv != jv:
               # Then create another edge which points in the reverse direction
                # <--this increments self.ne
                Dgraph.AddEdge(self, jv, iv, attr)
                self.ied_to_ieu[ied_redundant] = ie
                ied_redundant += 1

        self.neu = neu
        assert(self.ne == ned)

    def AddEdge(self, iv, jv, attr=None, remove_duplicates=False):
        """
        Add an edge to an undirected graph connecting vertices iv and jv.
        If the edge is already present (and remove_duplicates==True),
        no new edge will be added.

        Note: Undirected Ugraphs are implemented by creating two separate
              digraph edges that conect iv->jv  and jv->iv.

        """

        self.ieu_to_ied.append(len(self.edges))
        Dgraph.AddEdge(self, iv, jv, attr, remove_duplicates)
        self.ied_to_ieu.append(self.neu)
        if jv != iv:
            Dgraph.AddEdge(self, jv, iv, attr, remove_duplicates)
            self.ied_to_ieu.append(self.neu)
        self.neu += 1

        assert(len(self.ieu_to_ied) == self.neu)
        assert(len(self.ied_to_ieu) == len(self.edges))

    def ReorderEdges(self, epermutation, invert=False):
        Dgraph.ReorderEdges(self, epermutation, invert)

        # Now update the
        # self.ieu_to_ied and
        # self.ied_to_ieu lookup tables:

        if (invert):  # (first invert the permutation if necessary)
            eperm = [-1 for ie in epermutation]
            for ie in range(0, self.ne):
                eperm[epermutation[ie]] = ie
        else:
            eperm = epermutation
        # epermutation.reverse()

        ieu_to_ied_orig = [ied for ied in self.ieu_to_ied]
        ied_to_ieu_orig = [ieu for ieu in self.ied_to_ieu]

        for ieu in range(0, self.neu):
            ied_old = ieu_to_ied_orig[ieu]
            ied = eperm[ied_old]
            self.ieu_to_ied[ieu] = ied
        for ied_old in range(0, self.ne):
            ieu = ied_to_ieu_orig[ied_old]
            ied = eperm[ied_old]
            self.ied_to_ieu[ied] = ieu

        eperm = epermutation

    def LookupDirectedEdgeIdx(self, ieu):
        return self.ieu_to_ied[ieu]

    def LookupUndirectedEdgeIdx(self, ied):
        return self.ied_to_ieu[ied]

    # def GetVert(self, iv):    <-- (inherited from parent)
    #    return self.verts[iv]

    def GetEdge(self, ieu):
        ied = self.ieu_to_ied[ieu]
        return self.edges[ied]

    # def GetNumVerts(self):    <-- (inherited from parent)
    #    return self.nv

    def GetNumEdges(self):
        return self.neu

    def FindEdge(self, istart, istop):
        """
        A simple function looks up the (undirected) edge id number
        corresponding to an edge connecting vertices istart and istop.
        If not present returns Dgraph.NULL.

        To find the corresponding entry in the self.edges[] list,
        you can either:
            use the LookupDirectedEdge() lookup function
             or
            you can use the parent-class' version of this function
            Dgraph.FindEdge(self, istart, istop) which returns
            this number by default.

        """
        ied = Dgraph.FindEdge(self, istart, istop)
        ieu = self.LookupUndirectedEdgeIdx(ied)
        return ieu

    def CalcEdgeLookupTable(self):
        """
        COMMENT: THIS NEXT FUNCTION IS PROBABLY NOT NECESSARY AND MIGHT BE
                 REMOVED AT A LATER TIME WHEN I FIGURE OUT A BETTER WAY.

        Because undirected graphs (Ugraphs) are implemented as directed graphs
        (Dgraphs) with redundant edges, they may have some extra edges which
        the user never explicitly asked for.
        There is some confusion about whether the i'th edge refers to
        the i'th undirected edge that the user explicitly added, or
        the i'th directed edge which is stored internally.

           (The number of directed edges is usually twice the number of
           edges that the user asked for.  But not always, because edges
           wich start and end at the same vertex are only represented once.)

        This function calculates lookup tables to translate between
        the two edge numbering systems:

        self.ieu_to_ied[ieu] returns a directed edge id number,
                             (which is an index into the self.edges list)
                             corresponding to the ieu'th undirected edge
                             which was explicitly added by the caller.

        self.ied_to_ieu[ied] takes a directed edge id number (ied,
                             an index into the self.edges list)
                             and returns the undirected edge number,
                             which is allways <= ied

        """

        self.ieu_to_ied = []
        self.ied_to_ieu = [Ugraph.NULL for ied in range(0, self.ne)]
        for ied in range(0, self.ne):
            iv = self.edges[ied].start
            jv = self.edges[ied].stop
            ieu = len(self.ieu_to_ied)
            self.ied_to_ieu[ied] = ieu
            if iv <= jv:
                self.ieu_to_ied.append(ied)


def SortVertsByDegree(g):
    vert_numneighbors = [(iv, len(g.neighbors[iv])) for iv in range(0, g.nv)]
    vert_numneighbors.sort(key=itemgetter(1))
    order = [vert_numneighbors[iv][0] for iv in range(0, g.nv)]
    g.ReorderVerts(order, invert=True)


class DFS(object):
    """
    This class contains a member function (Order()) calculates the order
    of vertices visited in a depth-first-search over a connected graph.

    """

    def __init__(self, g):
        self.g = g
        self.sv = 0  # integer sv keeps track of how many vertices visited so far
        self.se = 0  # integer se keeps track of how many edges visited so far
        self.vvisited = [False for iv in range(0, self.g.nv)]  # verts visited
        self.vorder = [Dgraph.NULL for iv in range(
            0, self.g.nv)]  # search order
        self.evisited = [False for ie in range(0, self.g.ne)]  # edges visited
        self.eorder = [Dgraph.NULL for ie in range(
            0, self.g.ne)]  # search order

    def Reset(self):
        self.sv = 0
        self.se = 0
        for iv in range(0, self.g.nv):
            self.vvisited[iv] = False
            self.vorder[iv] = Dgraph.NULL
        for ie in range(0, self.g.ne):
            self.evisited[ie] = False
            self.eorder[ie] = Dgraph.NULL

    def Order(self, starting_node=0):
        """
        VisitOrder(starting_node)
        generates a list of integers from 0 to self.g.nv-1 (=#vertices minus 1)
        which represents the order in which the vertices would be visited
        during a Depth-First-Search.

        The first vertex visited is specified by the "starting_node" argument
        (an integer (from 0 to g.nv-1)).

        """
        self.Reset()
        # The first vertex to be visited should be the starting_node
        self.vorder[0] = starting_node
        self.vvisited[starting_node] = True
        self.sv = 1
        self._Order(starting_node)
        if self.sv != self.g.nv:
            raise(Disconnected(self.g, "Error(Order): " +
                               "The input graph is not connected."))
        assert(self.se == self.g.ne)
        return ([iv for iv in self.vorder], [ie for ie in self.eorder])
        # return self.order

    def _Order(self, iv):
        """
        _Order() is a recursive function which carries out a
        Depth-First-Search over the graph "self.g", starting with vertex iv.

        """
        for je in self.g.neighbors[iv]:
            jv = self.g.edges[je].stop
            if not self.evisited[je]:
                self.eorder[self.se] = je
                self.se += 1
                self.evisited[je] = True
                if not self.vvisited[jv]:
                    self.vorder[self.sv] = jv
                    self.sv += 1
                    self.vvisited[jv] = True
                    self._Order(jv)

    def IsConnected(self):
        self.Reset()
        self._Order(0)
        return (self.sv == self.g.nv)

    def IsCyclic(self):
        """
        IsCyclic() returns True if the graph is cyclic (and connected).
        (An exception is raised on disconnected graphs.)
        This function quits early as soon as a cycle is found.

        """
        self.Reset()
        if (type(self.g) is Ugraph):
            is_cyclic = self._IsCyclicUgraph(0, Dgraph.NULL)
        else:
            is_cyclic = self._IsCyclic(0)
        if ((self.sv != self.g.nv) and (not is_cyclic)):
            raise(Disconnected(self.g, "Error(IsCyclic): " +
                               "The input graph is not connected."))
        return is_cyclic

    def _IsCyclicUgraph(self, iv, ivprev):
        """
        _IsCyclicUgraph() is a recursive function which carries out a
        Depth-First-Search over the graph "self.g" to determine whether the
        graph is cyclic.  This function works on undirected graphs (Ugraphs).

        Indirected graphs (Ugraphs) are a special case.
        Ugraphs are implemented by using two (redundant) forward/backward edges
        connecting each pair of adjacent vertices.  This creates trivial loops.
        This version of _IsCyclicUgraph() only counts loops between more
        distantly connected vertices.

        """
        self.sv += 1
        self.vvisited[iv] = True
        for je in self.g.neighbors[iv]:
            jv = self.g.edges[je].stop
            if self.vvisited[jv]:
                if jv != ivprev:
                    return True
            elif self._IsCyclicUgraph(jv, iv):
                return True

        return False

    def _IsCyclic(self, iv):
        """
        _IsCyclic() is a recursive function which carries out a
        Depth-First-Search over the graph "self.g" to determine whether
        the graph is cyclic.
        This function works on directed graphs.

        """
        self.sv += 1
        self.vvisited[iv] = True
        for je in self.g.neighbors[iv]:
            jv = self.g.edges[je].stop
            if self.vvisited[jv]:
                return True
            elif self._IsCyclic(jv):
                return True

        return False


class GraphMatcher(object):
    """
    This class is a variant of the VF2 algorithm for searching
    for small connected subgraphs (g) within a larger graph (G).
    GraphMatcher works on directed or underected graphs (Dgraph or Ugraph).
    This particular version is better optimized for detecting subgraph
    isomorphisms between two graphs of highly unequal size.  It should be
    faster in these situations because, the computation required for
    each step is independent of the number of vertices in the larger graph
    In the original VF2 algorithm, the computation time for each step
    is proportional to the number of vertices in the larger graph.
    (The distinction matters when one graph is much smaller than the other.)

    Limitations: At the moment, the matching process uses a simple
    depth-first-search to search the vertices of the small graph "g".
    Hence this approach fails when the smaller graph g is disconnected.
    (but it can probably be fixed by picking a different algorithm to search
     the small graph).

    """

    def __init__(self,
                 G,  # The "big" graph
                 g):  # The little graph (number of vertices in g must be <= G)

        self.G = G
        self.g = copy.deepcopy(g)

        if (type(self.G) is Ugraph):
            assert(type(self.g) is Ugraph)
        #    self.G.CalcEdgeLookupTable() <-- not needed anymore

        self.sv = 0
        self.se = 0
        self.voccupiedG = [False for iv in range(0, G.nv)]
        self.eoccupiedG = [False for ie in range(0, G.ne)]
        self.G_is_too_small = False
        if ((g.nv > G.nv) or
                (g.ne > G.ne)):
            self.G_is_too_small = True
            # raise GenErr('Error: The first argument of GraphMatcher(G,g),\n'+
            #             '       must be at least as large as the second.')

        # The list self.iv_to_Iv is the mapping between the graph vertices.
        # Iv is an index into the large graph's list of vertices.
        # iv is an index into the small graph's list of vertices.
        # The mapping is stored in the iv_to_Iv list.
        self.iv_to_Iv = [Dgraph.NULL for Iv in range(0, self.g.nv)]
        self.ie_to_Ie = [Dgraph.NULL for Ie in range(0, self.g.ne)]
        #  (This used to be called "core_2" in the VF2 algorithm)

        # Due to the large number of recursion limit
        self.old_recursion_limit = sys.getrecursionlimit()
        expected_max_recursion = self.g.nv

        if self.old_recursion_limit < 1.5 * expected_max_recursion:
            # Give some breathing room.
            sys.setrecursionlimit(int(1.5 * expected_max_recursion))

        subgraph_searcher = DFS(self.g)
        # Perform a Depth-First-Search on the small graph.
        self.vorder_g, self.eorder_g = subgraph_searcher.Order()
        # Then re-order the vertices and edgers to
        # match the order they were visited.

        # Note on permutation order:
        # (The DFS.Order() function returns the permutation in this format
        #       old_index[ new_index ]
        #  where new_index is the DFS iteration when the vertex/edge was visited
        #    and old_index is the original vertex/edge order.
        #  However the ReorderVerts() and ReorderEdges() functions expect
        #  the permutation to have the opposite order: new_index[ old_index ]
        #  Hence we set "invert=True", when we invoke these functions.)
        self.g.ReorderVerts(self.vorder_g, invert=True)
        self.g.ReorderEdges(self.eorder_g, invert=True)

        # Initialize state
        self.Reset()

    def Reset(self):
        """Reinitializes the state of the match-search algorithm.

        """
        for iv in range(0, self.g.nv):
            self.iv_to_Iv[iv] = Dgraph.NULL
        for ie in range(0, self.g.ne):
            self.ie_to_Ie[ie] = Dgraph.NULL
        for Iv in range(0, self.G.nv):
            self.voccupiedG[Iv] = False
        for Ie in range(0, self.G.ne):
            self.eoccupiedG[Ie] = False

        self.se = 0
        self.sv = 0

        # OPTIONAL: First, do a partial sort for the vertices in the graphs
        # based on number of edges emanating from each vertex.
        # (This is probably unnecessary for small subgraphs.)
        # SortVertsByDegree(self.g)

    def Matches(self):
        """
        Iterator over all matches between G and g.
        Each "match" corresponds to a subgraph of G which is isomorphic to g.
        Matches is formatted as a 2-tuple of lists:
           (list of vertex ids from G, list of edge ids from G)
        The vertex ids in the list are a subset of the integers from 0 to G.nv.
        The edge   ids in the list are a subset of the integers from 0 to G.ne.

        (The corresponding vertices and edges from g are indicated by the order)

        """

        self.Reset()
        if self.G_is_too_small:
            # Then there are fewer verts and edges in G than in g.
            # Thus it is impossible for a subgraph of G to be isomorphic to g.
            return  # return no matches

        for Iv in range(0, self.G.nv):

            # match vertex Iv from G with vertex 0 from graph g
            self.iv_to_Iv[0] = Iv
            self.voccupiedG[Iv] = True

            # Implementation:
            # In this loop we begin the search process
            # starting with a different vertex (Iv) from big graph G,
            # and matching it with the first vertex (iv=0) from small graph g.
            # In this way the match "begins" from vertex Iv in G.
            #
            # Any matches found which begin from vertex Iv are distinct
            # from matches beginning from any other vertex in G.
            # Looping over all Iv in G is necessary and sufficient
            # to insure that all possible subgraphs of G
            # (which are isomorphic to g) are considered.

            self.sv = 1  # we have matched one vertex already
            self.se = 0  # we haven't matched any edges yet
            for match in self.Match():
                yield match
            self.voccupiedG[Iv] = False

    def Match(self):

        # self.se represents how many vertices have been matched so far.
        # We are done searching if all of the edges from 0 to self.se-1
        # from graph g have been selected (matched with edges from graph G).
        if self.se == self.g.ne:
            # Note: This also gaurantees that all vertices have been visited.
            assert(self.sv == self.g.nv)
            yield self.ReformatMatch()

        else:
            # VF2-style recursive loop:

            # We know the next edge to be matched is connected to at least
            # one previously visited vertex from g which has already been
            # been added to the the current match-in-progress.
            iv = self.g.edges[self.se].start
            Iv = self.iv_to_Iv[iv]
            assert(iv < self.sv)  # <-- check to verify this is so

            # The other vertex may or may not have been visited (matched) yet.
            iv_neighbor = self.g.edges[self.se].stop

            # Two cases:
            # Case 1: edge self.se points to a previously visited vertex from g
            #         This means we have a loop.
            if iv_neighbor < self.sv:
                # In that case, then the corresponding edge in G must
                # connect the corresponding pair of vertices from G.
                # (Which we know have already been assigned to vertices in g
                #  because both iv and iv_neighbor are < self.sv)
                Iv_neighbor = self.iv_to_Iv[iv_neighbor]
                # Loop over all of the edges in G which connect this pair
                # of vertices (Iv --> Iv_neighbor)
                for Je in self.G.neighbors[Iv]:
                    Jv = self.G.edges[Je].stop
                    if ((Jv == Iv_neighbor) and
                            (not self.eoccupiedG[Je])):

                        # Match edge Je from big   graph G with
                        #  edge self.se from small graph g
                        self.ie_to_Ie[self.se] = Je
                        self.se += 1
                        self.eoccupiedG[Je] = True
                        for match in self.Match():
                            yield match
                        self.eoccupiedG[Je] = False
                        self.se -= 1
                        self.ie_to_Ie[self.se] = Dgraph.NULL

            # Case 2:
            else:  # this would mean that iv_neighbor >= self.sv

                # If iv_neighbor>=self.sv, then this edge points to to a vertex
                # in g which has not yet been paired with a vertex from G.

                # Loop over all of the edges in G which connect vertex
                # Iv from G to new (unvisited) vertices in G
                for Je in self.G.neighbors[Iv]:
                    Jv = self.G.edges[Je].stop
                    if (not self.voccupiedG[Jv]):

                        assert(not self.eoccupiedG[Je])
                        # Match both edge Je with je
                        #      AND vertex Jv with jv
                        self.ie_to_Ie[self.se] = Je
                        self.se += 1
                        self.eoccupiedG[Je] = True
                        self.iv_to_Iv[self.sv] = Jv
                        self.sv += 1
                        self.voccupiedG[Jv] = True
                        # Then continue the recursion
                        for match in self.Match():
                            yield match
                        self.voccupiedG[Jv] = False
                        self.sv -= 1
                        self.iv_to_Iv[self.sv] = Dgraph.NULL
                        self.eoccupiedG[Je] = False
                        self.se -= 1
                        self.ie_to_Ie[self.se] = Dgraph.NULL

    def ReformatMatch(self):
        #   (This is because we are assuming g is connected.
        #    IT should not have any orphanned vertices.)
        # Now return the match:
        #
        # There are different ways of doing this
        # version 1:
        # match = (self.iv_to_Iv, self.ie_to_Ie) <-return a pointer to array
        # version 2:
        # match = ([Iv for Iv in self.iv_to_Iv], <-return a copy of the array
        #         [Ie for Ie in self.ie_to_Ie])
        # version 3:
        #   Recall that the vertices and edges and g have been re-ordered,
        #   so sort the list of Iv indices in the order they would be
        #   matched with the original vertices from the original graph g:
        # match = ([self.iv_to_Iv[self.vorder_g[iv]]
        #          for iv in range(0,self.g.nv)],
        #         [self.ie_to_Ie[self.eorder_g[ie]]
        #          for ie in range(0,self.g.ne)])
        # version 4: Similar to version 3 above, but we also translate
        #            the directed edge id list into a shorter undirected
        #            edge id list.
        match_verts = [self.iv_to_Iv[self.vorder_g[iv]]
                       for iv in range(0, self.g.nv)]

        if type(self.g) is Dgraph:
            match_edges = [self.ie_to_Ie[self.eorder_g[ie]]
                           for ie in range(0, self.g.ne)]
        else:
            #assert(atype(self.g) is Ugraph)
            match_edges = [Dgraph.NULL for ieu in range(0, self.g.neu)]

            for ie in range(0, self.g.ne):
                iv = self.g.edges[ie].start
                jv = self.g.edges[ie].stop
                if iv <= jv:  # <-- avoid duplicating edges (iv,jv) and (jv,iv)
                    ieu = self.g.LookupUndirectedEdgeIdx(ie)
                    Ie = self.ie_to_Ie[ie]
                    Ieu = self.G.LookupUndirectedEdgeIdx(Ie)
                    match_edges[ieu] = Ieu

        return (tuple(match_verts), tuple(match_edges))



# -*- coding: iso-8859-1 -*-
### -*- coding: utf-8 -*-
# Author: Andrew Jewett (jewett.aij at g mail)
#         http://www.chem.ucsb.edu/~sheagroup
# License: 3-clause BSD License  (See LICENSE.TXT)
# Copyright (c) 2012, Regents of the University of California
# All rights reserved.

"""A lexical analyzer class for simple shell-like syntaxes.
   This version has been modified slightly to work better with unicode.
   It was forked from the version of shlex that ships with python 3.2.2.
   A few minor features and functions have been added. """

# Module and documentation by Eric S. Raymond, 21 Dec 1998
# Input stacking and error message cleanup added by ESR, March 2000
# push_source() and pop_source() made explicit by ESR, January 2001.
# Posix compliance, split(), string arguments, and
# iterator interface by Gustavo Niemeyer, April 2003.
# ("wordterminators" (unicode support) hack by Andrew Jewett September 2011)

import os.path
import sys
from collections import deque
import re
import fnmatch
import string
#import gc


try:
    from cStringIO import StringIO
except ImportError:
    try:
        from StringIO import StringIO
    except ImportError:
        from io import StringIO

__all__ = ["TtreeShlex",
           "split",
           "LineLex",
           "SplitQuotedString",
           "EscCharStrToChar",
           "SafelyEncodeString",
           "RemoveOuterQuotes",
           "MaxLenStr",
           "HasWildcard",
           "MatchesPattern",
           #"IsRegex",
           "InputError",
           "ErrorLeader",
           "SrcLoc",
           "OSrcLoc",
           "TextBlock",
           "VarRef",
           "VarNPtr",
           "VarBinding",
           "SplitTemplate",
           "SplitTemplateMulti",
           "TableFromTemplate",
           "ExtractCatName",
           #"_TableFromTemplate",
           #"_DeleteLineFromTemplate",
           "DeleteLinesWithBadVars",
           "TemplateLexer"]


class TtreeShlex(object):
    """ A lexical analyzer class for simple shell-like syntaxes.
    TtreeShlex is a backwards-compatible version of python's standard shlex
    module. It has the additional member: "self.wordterminators", which
    overrides the "self.wordchars" member.  This enables better handling of
    unicode characters by allowing a much larger variety of characters to
    appear in words or tokens parsed by TtreeShlex.

    """

    def __init__(self,
                 instream=None,
                 infile=None,
                 posix=False):
        if isinstance(instream, str):
            instream = StringIO(instream)
        if instream is not None:
            self.instream = instream
            self.infile = infile
        else:
            self.instream = sys.stdin
            self.infile = None
        self.posix = posix
        if posix:
            self.eof = None
        else:
            self.eof = ''
        self.commenters = '#'
        self.wordchars = ('abcdfeghijklmnopqrstuvwxyz'
                          'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_')
        #if self.posix:
        #    self.wordchars += ('ßàáâãäåæçèéêëìíîïðñòóôõöøùúûüýþÿ'
        #                       'ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏÐÑÒÓÔÕÖØÙÚÛÜÝÞ')

        if self.posix:
            self.wordchars += ('ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½'
                               'ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½ï¿½')

        self.wordterminators = set([])
        self.prev_space_terminator = ''
        self.whitespace = ' \t\r\f\n'
        self.whitespace_split = False
        self.quotes = '\'"'
        self.escape = '\\'
        self.escapedquotes = '"'
        self.operators = '='
        self.state = ' '
        self.pushback = deque()
        self.lineno = 1
        self.debug = 0
        self.token = ''
        self.filestack = deque()
        # self.source_triggers
        # are tokens which allow the seamless insertion of other
        # files into the file being read.
        self.source_triggers = set(['source'])
        self.source_triggers_x = set([])
        # self.source_triggers_x is a subset of self.source_triggers.
        # In this case file inclusion is exclusive.
        # In other words, the file is only included
        # if it has not been included already.  It does this
        # by checking if one of these tokens has been encountered.
        self.source_files_restricted = set([])
        self.include_path = []
        if 'TTREE_PATH' in os.environ:
            include_path_list = os.environ['TTREE_PATH'].split(':')
            self.include_path += [d for d in include_path_list if len(d) > 0]
        if self.debug:
            sys.stderr.write('TtreeShlex: reading from %s, line %d'
                  % (self.instream, self.lineno))
        self.end_encountered = False

    @staticmethod
    def _belongs_to(char, include_chars, exclude_chars):
        if ((not exclude_chars) or (len(exclude_chars)==0)):
            return char in include_chars
        else:
            return char not in exclude_chars

    def push_raw_text(self, text):
        """Push a block of text onto the stack popped by the ReadLine() method.
        (If multiple lines are present in the text, (which is determined by
        self.line_terminators), then the text is split into multiple lines
        and each one of them is pushed onto this stack individually.
        The "self.lineno" counter is also adjusted, depending on the number
        of newline characters in "line".
            Do not strip off the newline, or other line terminators
            at the end of the text block before using push_raw_text()!

        """
        if self.debug >= 1:
            sys.stderr.write("TtreeShlex: pushing token " + repr(text))
        for c in reversed(text):
            self.pushback.appendleft(c)
            if c == '\n':
                self.lineno -= 1
        if len(text) > 0:
            self.end_encountered = False

    def push_token(self, text):
        "Push a token onto the stack popped by the get_token method"
        self.push_raw_text(text + self.prev_space_terminator)

    def push_source(self, newstream, newfile=None):
        "Push an input source onto the lexer's input source stack."
        if isinstance(newstream, str):
            newstream = StringIO(newstream)
        self.filestack.appendleft((self.infile, self.instream, self.lineno))
        self.infile = newfile
        self.instream = newstream
        self.lineno = 1
        if self.debug:
            if newfile is not None:
                sys.stderr.write('TtreeShlex: pushing to file %s' % (self.infile,))
            else:
                sys.stderr.write('TtreeShlex: pushing to stream %s' % (self.instream,))

    def pop_source(self):
        "Pop the input source stack."
        self.instream.close()
        (self.infile, self.instream, self.lineno) = self.filestack.popleft()
        if self.debug:
            sys.stderr.write('TtreeShlex: popping to %s, line %d'
                  % (self.instream, self.lineno))
        self.state = ' '

    def get_token(self):
        "Get a token from the input stream (or from stack if it's nonempty)"
        #### #CHANGING: self.pushback is now a stack of characters, not tokens
        #### if self.pushback:
        ####    tok = self.pushback.popleft()
        ####    if self.debug >= 1:
        ####        sys.stderr.write("TtreeShlex: popping token " + repr(tok))
        ####    return tok
        #### No pushback.  Get a token.
        raw = self.read_token()
        # Handle inclusions
        if self.source_triggers is not None:
            while raw in self.source_triggers:
                fname = self.read_token()
                spec = self.sourcehook(fname)
                if spec:
                    (newfile, newstream) = spec
                    if ((raw not in self.source_triggers_x) or
                            (newfile not in self.source_files_restricted)):
                        self.push_source(newstream, newfile)
                        if raw in self.source_triggers_x:
                            self.source_files_restricted.add(newfile)
                    else:
                        if self.debug >= 1:
                            sys.stderr.write(
                                '\ndebug warning: duplicate attempt to import file:\n               \"' + newfile + '\"\n')
                raw = self.get_token()

        # Maybe we got EOF instead?
        while raw == self.eof:
            if not self.filestack:
                return self.eof
            else:
                self.pop_source()
                raw = self.get_token()
        # Neither inclusion nor EOF
        if self.debug >= 1:
            if raw != self.eof:
                sys.stderr.write("TtreeShlex: token=" + repr(raw))
            else:
                sys.stderr.write("TtreeShlex: token=EOF")

        if raw == self.eof:
            self.end_encountered = True

        return raw

    def read_char(self):
        if self.pushback:
            nextchar = self.pushback.popleft()
            assert((type(nextchar) is str) and (len(nextchar)==1))
        else:
            nextchar = self.instream.read(1)
        return nextchar

    def read_token(self):
        self.prev_space_terminator = ''
        quoted = False
        escapedstate = ' '
        while True:
            #### self.pushback is now a stack of characters, not tokens
            nextchar = self.read_char()
            if nextchar == '\n':
                self.lineno = self.lineno + 1
            if self.debug >= 3:
                sys.stderr.write("TtreeShlex: in state", repr(self.state),
                      "I see character:", repr(nextchar))
            if self.state is None:
                self.token = ''        # past end of file
                break
            elif self.state == ' ':
                if not nextchar:
                    self.state = None  # end of file
                    break
                elif nextchar in self.whitespace:
                    if self.debug >= 2:
                        sys.stderr.write("TtreeShlex: I see whitespace in whitespace state")
                    if self.token or (self.posix and quoted):
                        # Keep track of which whitespace
                        # character terminated the token.
                        self.prev_space_terminator = nextchar
                        break   # emit current token
                    else:
                        continue
                elif nextchar in self.commenters:
                    self.instream.readline()
                    self.lineno = self.lineno + 1
                elif self.posix and nextchar in self.escape:
                    escapedstate = 'a'
                    self.state = nextchar
                elif TtreeShlex._belongs_to(nextchar,
                                            self.wordchars,
                                            self.wordterminators):
                    self.token = nextchar
                    self.state = 'a'
                elif nextchar in self.quotes:
                    if not self.posix:
                        self.token = nextchar
                    self.state = nextchar
                elif self.whitespace_split:
                    self.token = nextchar
                    self.state = 'a'
                else:
                    self.token = nextchar
                    if self.token or (self.posix and quoted):
                        break   # emit current token
                    else:
                        continue
            elif self.state in self.quotes:
                quoted = True
                if not nextchar:      # end of file
                    if self.debug >= 2:
                        sys.stderr.write("TtreeShlex: I see EOF in quotes state")
                    # XXX what error should be raised here?
                    raise ValueError("Error at or before " + self.error_leader() + "\n"
                                     "      No closing quotation.")
                if nextchar == self.state:
                    if not self.posix:
                        self.token = self.token + nextchar
                        self.state = ' '
                        break
                    else:
                        self.state = 'a'
                elif self.posix and nextchar in self.escape and \
                        self.state in self.escapedquotes:
                    escapedstate = self.state
                    self.state = nextchar
                else:
                    self.token = self.token + nextchar
            elif self.state in self.escape:
                if not nextchar:      # end of file
                    if self.debug >= 2:
                        sys.stderr.write("TtreeShlex: I see EOF in escape state")
                    # XXX what error should be raised here?
                    raise ValueError("No escaped character")
                # In posix shells, only the quote itself or the escape
                # character may be escaped within quotes.
                if escapedstate in self.quotes and \
                   nextchar != self.state and nextchar != escapedstate:
                    self.token = self.token + self.state
                self.token = self.token + nextchar
                self.state = escapedstate
            elif self.state == 'a':
                if not nextchar:
                    self.state = None   # end of file
                    break
                elif nextchar in self.whitespace:
                    if self.debug >= 2:
                        sys.stderr.write("TtreeShlex: I see whitespace in word state")
                    self.state = ' '
                    if self.token or (self.posix and quoted):
                        # Keep track of which whitespace
                        # character terminated the token.
                        self.prev_space_terminator = nextchar
                        break   # emit current token
                    else:
                        continue
                elif nextchar in self.commenters:
                    comment_contents = self.instream.readline()
                    self.lineno = self.lineno + 1
                    if self.posix:
                        self.state = ' '
                        if self.token or (self.posix and quoted):
                            # Keep track of which character(s) terminated
                            # the token (including whitespace and comments).
                            self.prev_space_terminator = nextchar + comment_contents
                            break   # emit current token
                        else:
                            continue
                elif self.posix and nextchar in self.quotes:
                    self.state = nextchar
                elif self.posix and nextchar in self.escape:
                    escapedstate = 'a'
                    self.state = nextchar
                elif (TtreeShlex._belongs_to(nextchar,
                                             self.wordchars,
                                             self.wordterminators)
                      or (nextchar in self.quotes)
                      or (self.whitespace_split)):
                    self.token = self.token + nextchar
                else:
                    self.pushback.appendleft(nextchar)
                    if self.debug >= 2:
                        sys.stderr.write("TtreeShlex: I see punctuation in word state")
                    self.state = ' '
                    if self.token:
                        break   # emit current token
                    else:
                        continue
        result = self.token
        self.token = ''
        if self.posix and not quoted and result == '':
            result = None
        if self.debug > 1:
            if result:
                sys.stderr.write("TtreeShlex: raw token=" + repr(result))
            else:
                sys.stderr.write("TtreeShlex: raw token=EOF")
        return result

    def sourcehook(self, newfile):
        "Hook called on a filename to be sourced."
        newfile = RemoveOuterQuotes(newfile)
        # This implements cpp-like semantics for relative-path inclusion.
        if isinstance(self.infile, str) and not os.path.isabs(newfile):
            newfile_full = os.path.join(os.path.dirname(self.infile), newfile)
        try:
            f = open(newfile_full, "r")
        except IOError:
            # If not found,
            err = True
            # ...then check to see if the file is in one of the
            # directories in the self.include_path list.
            for d in self.include_path:
                newfile_full = os.path.join(d, newfile)
                try:
                    f = open(newfile_full, "r")
                    err = False
                    break
                except IOError:
                    err = True
            if err:
                raise InputError('Error at ' + self.error_leader() + '\n'
                                 '       unable to open file \"' + newfile + '\"\n'
                                 '       for reading.\n')
        return (newfile, f)

    def error_leader(self, infile=None, lineno=None):
        "Emit a C-compiler-like, Emacs-friendly error-message leader."
        if infile is None:
            infile = self.infile
        if lineno is None:
            lineno = self.lineno
        return "\"%s\", line %d: " % (infile, lineno)

    def __iter__(self):
        return self

    def __next__(self):
        token = self.get_token()
        if token == self.eof:
            raise StopIteration
        return token

    def __bool__(self):
        return not self.end_encountered

    # For compatibility with python 2.x, I must also define:
    def __nonzero__(self):
        return self.__bool__()


# The split() function was originally from shlex
# It is included for backwards compatibility.
def split(s, comments=False, posix=True):
    lex = TtreeShlex(s, posix=posix)
    lex.whitespace_split = True
    if not comments:
        lex.commenters = ''
    return list(lex)


##################### NEW ADDITIONS (may be removed later) #################

#"""
#  -- linelex.py --
# linelex.py defines the LineLex class, which inherits from, and further
# augments the capabilities of TtreeShlex by making it easier to parse
# individual lines one at a time.  (The original shlex's "source" inclusion
# ability still works when reading entire lines, and lines are still counted.)
#
#"""

#import sys


class InputError(Exception):
    """ A generic exception object containing a string for error reporting.
        (Raising this exception implies that the caller has provided
         a faulty input file or argument.)

    """

    def __init__(self, err_msg):
        self.err_msg = err_msg

    def __str__(self):
        return self.err_msg

    def __repr__(self):
        return str(self)


def ErrorLeader(infile, lineno):
    return '\"' + infile + '\", line ' + str(lineno)


class SrcLoc(object):
    """ SrcLoc is essentially nothing more than a 2-tuple containing the name
    of a file (str) and a particular line number inside that file (an integer).

    """
    __slots__ = ["infile", "lineno"]

    def __init__(self, infile='', lineno=-1):
        self.infile = infile
        self.lineno = lineno


def SplitQuotedString(string,
                      quotes='\'\"',
                      delimiters=' \t\r\f\n',
                      escape='\\',
                      comment_char='#'):
    tokens = []
    token = ''
    reading_token = True
    escaped_state = False
    quote_state = None
    for c in string:

        if (c in comment_char) and (not escaped_state) and (quote_state == None):
            tokens.append(token)
            return tokens

        elif (c in delimiters) and (not escaped_state) and (quote_state == None):
            if reading_token:
                tokens.append(token)
                token = ''
                reading_token = False

        elif c in escape:
            if escaped_state:
                token += c
                reading_token = True
                escaped_state = False
            else:
                escaped_state = True
                # and leave c (the '\' character) out of token
        elif (c in quotes) and (not escaped_state):
            if (quote_state != None):
                if (c == quote_state):
                    quote_state = None
            else:
                quote_state = c
            token += c
            reading_token = True
        else:
            if (c == 'n') and (escaped_state == True):
                c = '\n'
            elif (c == 't') and (escaped_state == True):
                c = '\t'
            elif (c == 'r') and (escaped_state == True):
                c = '\r'
            elif (c == 'f') and (escaped_state == True):
                c = '\f'
            token += c
            reading_token = True
            escaped_state = False

    if len(string) > 0:
        tokens.append(token)
    return tokens


def EscCharStrToChar(s_in, escape='\\'):
    """
    EscCharStrToChar() replaces any escape sequences
    in a string with their 1-character equivalents.

    """
    assert(len(escape) > 0)
    out_lstr = []
    escaped_state = False
    for c in s_in:
        if escaped_state:
            if (c == 'n'):
                out_lstr.append('\n')
            elif (c == 't'):
                out_lstr.append('\t')
            elif (c == 'r'):
                out_lstr.append('\r')
            elif (c == 'f'):
                out_lstr.append('\f')
            elif (c == '\''):
                out_lstr.append('\'')
            elif (c == '\"'):
                out_lstr.append('\"')
            elif c in escape:
                out_lstr.append(c)
            else:
                out_lstr.append(escape + c)  # <- keep both characters
            escaped_state = False
        else:
            if c in escape:
                escaped_state = True
            else:
                out_lstr.append(c)

    return ''.join(out_lstr)


def SafelyEncodeString(in_str,
                       quotes='\'\"',
                       delimiters=' \t\r\f\n',
                       escape='\\',
                       comment_char='#'):
    """
    SafelyEncodeString(in_str) scans through the input string (in_str),
    and returns a new string in which probletic characters
    (like newlines, tabs, quotes, etc), are replaced by their two-character
    backslashed equivalents (like '\n', '\t', '\'', '\"', etc).
    The escape character is the backslash by default, but it too can be
    overridden to create custom escape sequences
    (but this does not effect the encoding for characters like '\n', '\t').

    """
    assert(len(escape) > 0)
    out_lstr = []
    use_outer_quotes = False
    for c in in_str:
        if (c == '\n'):
            c = '\\n'
        elif (c == '\t'):
            c = '\\t'
        elif (c == '\r'):
            c = '\\r'
        elif (c == '\f'):
            c = '\\f'
        elif c in quotes:
            c = escape[0] + c
        elif c in escape:
            c = c + c
        elif c in delimiters:
            use_outer_quotes = True
        # hmm... that's all that comes to mind.  Did I leave anything out?
        out_lstr.append(c)

    if use_outer_quotes:
        out_lstr = ['\"'] + out_lstr + ['\"']

    return ''.join(out_lstr)


def RemoveOuterQuotes(text, quotes='\"\''):
    if ((len(text) >= 2) and (text[0] in quotes) and (text[-1] == text[0])):
        return text[1:-1]
    else:
        return text


def MaxLenStr(s1, s2):
    if len(s2) > len(s1):
        return s2
    else:
        return s1


# def IsRegex(pat):
#    """
#    Check to see if string (pat) is bracketed by slashes.
#
#    """
#    return (len(pat)>=2) and (pat[0]=='/') and (pat[-1] == '/')

def HasWildcard(pat):
    """
    Returns true if a string (pat) contains a '*' or '?' character.

    """
    return (pat.find('*') != -1) or (pat.find('?') != -1)


# def HasWildcard(pat):
#    """
#    Returns true if a string (pat) contains a non-backslash-protected
#    * or ? character.
#
#    """
#    N=len(pat)
#    i=0
#    while i < N:
#        i = pat.find('*', i, N)
#        if i == -1:
#            break
#        elif (i==0) or (pat[i-1] != '\\'):
#            return True
#        i += 1
#    i=0
#    while i < N:
#        i = pat.find('?', i, N)
#        if i == -1:
#            break
#        elif (i==0) or (pat[i-1] != '\\'):
#            return True
#        i += 1
#    return False


def MatchesPattern(s, pattern):
    if type(pattern) is str:
        # old code:
        # if ((len(s) > 1) and (s[0] == '/') and (s[-1] == '/'):
        #    re_string = p[1:-1]  # strip off the slashes '/' and '/'
        #    if not re.search(re_string, s):
        #        return False
        # new code:
        #    uses precompiled regular expressions (See "pattern.search" below)
        if HasWildcard(pattern):
            if not fnmatch.fnmatchcase(s, pattern):
                return False
        elif s != pattern:
            return False
    else:
        #assert(type(p) is _sre.SRE_Match)
        # I assume pattern = re.compile(some_reg_expr)
        if not pattern.search(s):
            return False
    return True


def MatchesAll(multi_string, pattern):
    assert(len(multi_string) == len(pattern))
    for i in range(0, len(pattern)):
        if not MatchesPattern(multi_string[i], pattern[i]):
            return False
    return True


class LineLex(TtreeShlex):
    """ This class extends the TtreeShlex module (a slightly modified
    version of the python 3.2.2 version of shlex).  LineLex has the
    ability to read one line at a time (in addition to one token at a time).
    (Many files and scripts must be parsed one line at a time instead of one
     token at a time.  In these cases, the whitespace position also matters.)

    Arguably, this class might not be necessary.
    I could get rid of this class completely.  That would be nice.  To do that
    we would need to augment and generalize shlex's get_token() member function
    to make it read lines, not just tokens.  Of course, you can always
    change the wordchars (or wordterminators).  Even so, there are two other
    difficulties using the current version of shlex.get_token() to read lines:
    1) File inclusion happen whenever the beginning of a line/token matches one
       of the "source_triggers" (not the whole line as required by get_token()).
    2) Lines ending in a special character (by default the backslash character)
       continue on to the next line.
    This code seems to work on our test files, but I'm sure there are bugs.
    Andrew 2012-3-25

    """

    def __init__(self,
                 instream=None,
                 infile=None,
                 posix=False):
        TtreeShlex.__init__(self, instream, infile, posix)
        self.line_terminators = '\n'
        self.line_extend_chars = '\\'
        self.skip_comments_during_readline = True

    def _StripComments(self, line):
        if self.skip_comments_during_readline:
            for i in range(0, len(line)):
                if ((line[i] in self.commenters) and
                        ((i == 0) or (line[i - 1] not in self.escape))):
                    return line[:i]
        return line

    def _ReadLine(self,
                  recur_level=0):
        """
        This function retrieves a block of text, halting at a
        terminal character.  Escape sequences are respected.
        The self.lineno (newline counter) is also maintained.

        The main difference between Readline and get_token()
        is the way they handle the "self.source_triggers" member.
        Both Readline() and get_token() insert text from other files when they
        encounter a string in "self.source_triggers" in the text they read.
        However ReadLine() ONLY inserts text from other files if the token which
        matches with self.source_triggers appears at the beginning of the line.
        get_token() inserts text only if lex.source matches the entire token.

        comment-to-self:
         At some point, once I'm sure this code is working, I should replace
         shlex.get_token() with the code from ReadLine() which is more general.
         It would be nice to get rid of "class LineLex" entirely.  ReadLine()
         is the only new feature that LineLex which was lacking in shlex.

         To do this I would need to add a couple optional arguments to
         "get_token()", allowing it to mimic ReadLine(), such as:
           "override_wordterms" argument (which we can pass a '\n'), and
           "token_extender" argument (like '\' for extending lines)

        """
        first_token = ''
        line = ''
        escaped_state = False
        found_space = False
        while True:
            nextchar = self.read_char()
            # sys.stderr.write('nextchar=\"'+nextchar+'\"\n')
            while nextchar == '':
                if not self.filestack:
                    return self._StripComments(line), '', first_token, found_space
                else:
                    self.pop_source()
                    nextchar = self.read_char()
            if nextchar == '\n':
                self.lineno += 1

            if escaped_state:
                escaped_state = False
            else:
                if nextchar in self.escape:
                    line += nextchar
                    escaped_state = True
                else:
                    escaped_state = False

            if not escaped_state:
                if (nextchar in self.whitespace):
                    found_space = True
                    while first_token in self.source_triggers:
                        fname = RemoveOuterQuotes(self.get_token())
                        if (fname == '') or (fname in self.source_triggers):
                            raise InputError('Error: near ' + self.error_leader() + '\n'
                                             '       Nonsensical file inclusion request.\n')
                        if self.debug >= 0:
                            sys.stderr.write(('  ' * recur_level) +
                                             'reading file \"' + fname + '\"\n')
                        spec = self.sourcehook(fname)
                        if spec:
                            (fname, subfile) = spec
                            if ((first_token not in self.source_triggers_x) or
                                (fname not in self.source_files_restricted)):
                                self.push_source(subfile, fname)
                            if first_token in self.source_triggers_x:
                                self.source_files_restricted.add(fname)
                            else:
                                if self.debug >= 0:
                                    sys.stderr.write('\nWarning at ' + self.error_leader() + ':\n'
                                                     '          duplicate attempt to import file:\n         \"' + fname + '\"\n')

                        line, nextchar, first_token, found_space = \
                            self._ReadLine(recur_level + 1)

                if nextchar in self.line_terminators:
                    line_nrw = line.rstrip(self.whitespace)
                    # sys.stderr.write('line_nrw=\"'+line_nrw+'\"\n')
                    if ((len(line_nrw) > 0) and
                            (line_nrw[-1] in self.line_extend_chars) and
                            ((len(line_nrw) < 2) or (line_nrw[-2] not in self.escape))):
                        # delete the line_extend character
                        line = line_nrw[:-1]
                        # from the end of that line and keep reading...
                    else:
                        return self._StripComments(line), nextchar, first_token, found_space
                else:
                    line += nextchar
                    if not found_space:
                        first_token += nextchar

    def ReadLine(self, recur_level=0):
        line, nextchar, first_token, found_space = \
            self._ReadLine(recur_level)
        if nextchar == self.eof:
            self.end_encountered = True
        return line + nextchar

    @staticmethod
    def TextBlock2Lines(text, delimiters, keep_delim=True):
        """ This splits a string into a list of sub-strings split by delimiter
        characters.  This function is different from the standard str.split()
        function: The string is split at every character which belongs to the
        "delimiters" argument (which can be a string or some other container).
        This character is included at the end of every substring.  Example:
        TextBlock2Lines('\nabc\nde^fg\nhi j\n', '^\n')
        returns:
        ['\n', 'abc\n', 'de^', 'fg\n', 'hi j\n']

        """
        ls = []
        i = 0
        i_prev = 0
        while i < len(text):
            if text[i] in delimiters:
                if keep_delim:
                    ls.append(text[i_prev:i + 1])
                else:
                    ls.append(text[i_prev:i])
                i_prev = i + 1
            i += 1
        if (i_prev < len(text)):
            ls.append(text[i_prev:i + 1])
        return ls

    def __iter__(self):
        return self

    def __next__(self):
        line = self.ReadLine()
        if line == self.eof:
            raise StopIteration
        return line


class OSrcLoc(object):
    """ OSrcLoc is barely more than a 2-tuple containing the name of a file
        (a string) and a particular line number inside that file (an integer).
        These objects are passed around and stored in the nodes of
        every tree, so that if a syntax error or broken link in that node
        is discovered, an error message can be provided to the user.

    """

    __slots__ = ["infile", "lineno", "order"]
    count = 0

    def __init__(self, infile='', lineno=-1):
        self.infile = infile
        self.lineno = lineno
        OSrcLoc.count += 1
        self.order = OSrcLoc.count  # keep track of how many times it was called

    def __lt__(self, x):
        return self.order < x.order

    # def __repr__(self):
    #    return repr((self.infile, self.lineno, self.order))


class TextBlock(object):
    """TextBlock is just a 3-tuple consisting of a string, and an OSrcLoc
       to help locate it in the original file from which it was read."""

    __slots__ = ["text", "srcloc"]

    def __init__(self, text, srcloc):  # srcloc_end):
        self.text = text
        if srcloc == None:
            self.srcloc = OSrcLoc()
        else:
            self.srcloc = srcloc
        # if srcloc_end == None:
        #    self.srcloc_end = OSrcLoc()
        # else:
        #    self.srcloc_end = srcloc_end

    def __repr__(self):
        return '\"' + self.text + '\"'


class VarRef(object):
    """VarRef stores variable names, and paths, and other attribute information,
    as well as a "OSrcLoc" to keep track of the file it was defined in."""

    __slots__ = ["prefix", "descr_str", "suffix", "srcloc", "binding", "nptr"]

    def __init__(self,
                 prefix='',  # '$' or '${'
                 descr_str='',  # <- descriptor string: "cpath/category:lpath"
                 suffix='',  # '}'
                 srcloc=None,  # location in file where defined
                 binding=None,  # a pointer to a tuple storing the value
                 nptr=None):  # <- see class VarNPtr

        self.prefix = prefix  # Any text before the descriptor string goes here
        self.suffix = suffix  # Any text after the descriptor string goes here
        self.descr_str = descr_str
        if srcloc == None:  # <- Location in text file where variable appears
            self.srcloc = OSrcLoc()
        else:
            self.srcloc = srcloc

        self.binding = binding

        if nptr == None:
            self.nptr = VarNPtr()
        else:
            self.nptr = nptr

    def __lt__(self, x):
        return self.order < x.order

    # def __repr__(self):
    #    return repr((self.prefix + self.descr_str + self.suffix, srcloc))


class VarNPtr(object):
    """
    Every time a variable appears in a template, it has has a "descriptor".
    For example, consider the variable
       "$atom:CA"
    This is a string which encodes 3 pieces of information.
    1) the category name:  This is essentialy indicates the variable's type.
                           (ie "atom", in the example above)
    2) the category node:  Some TYPES have limited scope. Users can
                           specify the root node of the portion of the tree
                           in which this variable's type makes sense.
                           If this node is the root node, then that category
                           is relevant everywhere, and is not molecule or class
                           specific.  All variables have a category node, which
                           is often not explicitly defined to by the user.
                           (Category node = the root "/", in the example above.)
    3) the leaf node:      This is a node whose ".name" member matches the name
                           of a variable.  This node is created for this purpose
                           and it's position in the tree is a reflection of
                           that variable's intended scope.
                              In a molecule this "name" might be the name
                           of a type of atom, or an atom ID, or a bond type,
                           which is found in a particular molecule.
                           (Leaf node would be named "CA" in the example above.)

    The VarNPtr class is simply a 3-tuple which
    keeps these 3 pieces of data together.

    """

    __slots__ = ["cat_name", "cat_node", "leaf_node"]

    def __init__(self, cat_name='', cat_node=None, leaf_node=None):
        self.cat_name = cat_name
        self.cat_node = cat_node
        self.leaf_node = leaf_node

    # def __repr__(self):
    #    return repr((self.cat_name, self.cat_node.name, self.leaf_node.name))


class VarBinding(object):
    """ VarBinding is essentially a tuple consistng of (full_name, binding, refs):

    "self.full_name" is canonical name for this variable.  This is a string
    which specifies full path leading to the category node (beginning with '/'),
    the category name (followed by a ':'),
    as well as the leaf node (including the path leading up to it from cat_node)
    This triplet identifies the variable uniquely.

    "self.value" is the data that the variable refers to (usually a string).

    "self.refs" stores a list of VarRefs which mention the same variable
    from the various places inside various templates in the tree.

    """

    __slots__ = ["full_name", "nptr", "value", "refs", "order", "category"]

    def __init__(self,
                 full_name='',
                 nptr=None,
                 value=None,
                 refs=None,
                 order=-1,
                 category=None):
        self.full_name = full_name
        self.nptr = nptr
        self.value = value
        self.refs = refs
        self.order = order
        self.category = category

    def __lt__(self, x):
        return self.order < x.order

    def __repr__(self):
        return repr((self.full_name, self.value, self.order))


def ExtractCatName(descr_str):
    """ When applied to a VarRef's "descr_str" member,
    this function will extract the "catname" of it's corresponding
    "nptr" member.  This can be useful for error reporting.
    (I use it to insure that the user is using the correct counter
     variable types at various locations in their input files.)

    """

    ib = descr_str.find(':')
    if ib == -1:
        ib = len(descr_str)
        ia = descr_str.rfind('/')
        if ia == -1:
            ia = 0
        return descr_str[ia:ib]
    else:
        str_before_colon = descr_str[0:ib]
        ia = str_before_colon.rfind('/')
        if ia == -1:
            return str_before_colon
        else:
            return str_before_colon[ia + 1:]


def _DeleteLineFromTemplate(tmpl_list,
                            i_entry,  # index into tmpl_list
                            newline_delimiter='\n'):
    """ Delete a single line from tmpl_list.
    tmpl_list is an alternating list of VarRefs and TextBlocks.
    To identify the line, the index corresponding to one of the
    entries in the tmpl_list is used. (Usually it is a VarRef)
    The text after the preceeding newline, and the text up to the next newline
       (starting from the beginning of the current entry, if a TextBlock)
    is deleted, including any VarRef (variables) located in between.

    It returns the index corresponding to the next
    entry in the list (after deletion).

    """

    i_prev_newline = i_entry
    while i_prev_newline >= 0:
        entry = tmpl_list[i_prev_newline]
        if isinstance(entry, TextBlock):
            i_char_newline = entry.text.rfind(newline_delimiter)
            if i_char_newline != -1:  # then newline found
                # Delete the text after this newline
                entry.text = entry.text[:i_char_newline + 1]
                break
        i_prev_newline -= 1

    first_var = True
    #i_next_newline = i_entry
    i_next_newline = i_prev_newline + 1
    while i_next_newline < len(tmpl_list):
        entry = tmpl_list[i_next_newline]
        if isinstance(entry, TextBlock):
            i_char_newline = entry.text.find(newline_delimiter)
            if i_char_newline != -1:  # then newline found
                # Delete the text before this newline (including the newline)
                entry.text = entry.text[i_char_newline + 1:]
                break
        # Invoke DeleteSelf() on the first variables on this line.  This will
        # insure that it is deleted from the ttree_assignments.txt file.
        elif isinstance(entry, VarRef):
            if first_var:
                entry.nptr.leaf_node.DeleteSelf()
            first_var = False
        i_next_newline += 1

    del tmpl_list[i_prev_newline + 1: i_next_newline]
    return i_prev_newline + 1


def DeleteLinesWithBadVars(tmpl_list,
                           delete_entire_template=False,
                           newline_delimiter='\n'):
    """
    Loop through the entries in a template,
    an alternating list of TextBlocks and VarRefs (tmpl_list).
    If a VarRef points to a leaf_node which no longer exists
    (ie. no longer in the corresponding category's .bindings list).
    Then delete the line it came from from the template (tmpl_list).

    """

    out_str_list = []
    i = 0
    while i < len(tmpl_list):
        entry = tmpl_list[i]
        if isinstance(entry, VarRef):
            var_ref = entry
            var_bindings = var_ref.nptr.cat_node.categories[
                var_ref.nptr.cat_name].bindings
            # if var_ref.nptr.leaf_node not in var_bindings:
            if var_ref.nptr.leaf_node.IsDeleted():
                if delete_entire_template:
                    del tmpl_list[:]
                    return 0
                else:
                    i = _DeleteLineFromTemplate(tmpl_list,
                                                i,
                                                newline_delimiter)
            else:
                i += 1
        else:
            i += 1


def SplitTemplate(ltmpl, delim, delete_blanks=False):
    """
    Split a template "ltmpl" into a list of "tokens" (sub-templates)
    using a single delimiter string "delim".

    INPUT arguments:
    "ltmpl" should be an list of TextBlocks and VarRefs.
    "delim" should be a simple string (type str)
    "delete_blanks" should be a boolean True/False value.
                    When true, successive occurrences of the delimiter
                    should not create blank entries in the output list.

    OUTPUT:
    A list of tokens.
    Each "token" is either a TextBlock, a VarRef,
    or a (flat, 1-dimensional) list containing more than one of these objects.
    The number of "tokens" returned equals the number of times the delimiter
    is encountered in any of the TextBlocks in the "ltmpl" argument, plus one.
    (... Unless "delete_blanks" is set to True.
     Again, in that case, empty entries in this list are deleted.)

    """
    assert(type(delim) is str)
    if not hasattr(ltmpl, '__len__'):
        ltmpl = [ltmpl]

    tokens_lltmpl = []
    token_ltmpl = []
    i = 0
    while i < len(ltmpl):

        entry = ltmpl[i]
        #sys.stderr.write('ltmpl['+str(i)+'] = '+str(entry)+'\n')

        if isinstance(entry, TextBlock):
            # if hasattr(entry, 'text'):
            prev_src_loc = entry.srcloc

            tokens_str = entry.text.split(delim)

            lineno = entry.srcloc.lineno

            j = 0
            while j < len(tokens_str):
                token_str = tokens_str[j]

                delim_found = False
                if (j < len(tokens_str) - 1):
                    delim_found = True

                if token_str == '':
                    if delete_blanks:
                        if delim == '\n':
                            lineno += 1
                        if len(token_ltmpl) > 0:
                            if len(token_ltmpl) == 1:
                                tokens_lltmpl.append(token_ltmpl[0])
                            else:
                                tokens_lltmpl.append(token_ltmpl)
                        del token_ltmpl
                        token_ltmpl = []
                        j += 1
                        continue

                new_src_loc = OSrcLoc(prev_src_loc.infile, lineno)
                new_src_loc.order = prev_src_loc.order

                for c in token_str:
                    # Reminder to self:  c != delim  (so c!='\n' if delim='\n')
                    # (We keep track of '\n' characters in delimiters above.)
                    if c == '\n':
                        lineno += 1

                new_src_loc.lineno = lineno

                text_block = TextBlock(token_str,
                                       new_src_loc)

                prev_src_loc = new_src_loc

                if len(token_ltmpl) == 0:
                    if delim_found:
                        tokens_lltmpl.append(text_block)
                        del token_ltmpl
                        token_ltmpl = []
                    else:
                        token_ltmpl.append(text_block)
                else:
                    if delim_found:
                        if len(token_str) > 0:
                            token_ltmpl.append(text_block)
                            tokens_lltmpl.append(token_ltmpl)
                            del token_ltmpl
                            token_ltmpl = []
                        else:
                            assert(not delete_blanks)
                            if (isinstance(token_ltmpl[-1], VarRef)
                                and
                                ((j > 0)
                                 or
                                 ((j == len(tokens_str) - 1) and
                                  (i == len(ltmpl) - 1))
                                 )):
                                # In that case, this empty token_str corresponds
                                # to a delimiter which was located immediately
                                # after the variable name,
                                # AND
                                #   -there is more text to follow,
                                #   OR
                                #   -we are at the end of the template.
                                token_ltmpl.append(text_block)
                            if len(token_ltmpl) == 1:
                                tokens_lltmpl.append(token_ltmpl[0])
                            else:
                                tokens_lltmpl.append(token_ltmpl)
                            del token_ltmpl
                            token_ltmpl = []
                    else:
                        token_ltmpl.append(text_block)

                if (delim_found and (delim == '\n')):
                    lineno += 1

                j += 1

        elif isinstance(entry, VarRef):
            # elif hasattr(entry, 'descr_str'):
            lineno = entry.srcloc.lineno
            if ((len(token_ltmpl) == 1) and
                    isinstance(token_ltmpl[0], TextBlock) and
                    (len(token_ltmpl[0].text) == 0)):
                # special case: if the previous entry was "", then it means
                # the delimeter appeared at the end of the previous text block
                # leading up to this variable.  It separates the variable from
                # the previous text block.  It is not a text block of length 0.
                token_ltmpl[0] = entry
            else:
                token_ltmpl.append(entry)
        elif entry == None:
            token_ltmpl.append(entry)
        else:
            assert(False)

        i += 1

    # Append left over remains of the last token
    if len(token_ltmpl) == 1:
        tokens_lltmpl.append(token_ltmpl[0])
    elif len(token_ltmpl) > 1:
        tokens_lltmpl.append(token_ltmpl)
    del token_ltmpl

    return tokens_lltmpl


def SplitTemplateMulti(ltmpl, delims, delete_blanks=False):
    """
    Split a template "ltmpl" into a list of templates using a
    single one or more delimiter strings "delim_list".
    If multiple delimiter strings are provided, splitting
    begins using the first delimiter string in the list.
    Then each token in the resulting list of templates
    is split using the next delimiter string
    and so on until we run out of delimiter strings.

    "ltmpl" should be an list of TextBlocks and VarRefs.
    "delims" should be a simple string (type str) or a list of strings
    "delete_blanks" is either True or False
                    If True, then any blank entries in the resulting list of
                    tokens (sub-templates) will be deleted.

    """

    if hasattr(delims, '__len__'):  # then it hopefully is a list of strings
        delim_list = delims
    else:
        delim_list = [delims]     # then it hopefully is a string

    tokens = [ltmpl]
    for delim in delim_list:
        assert(type(delim) is str)
        tokens_il = []
        for t in tokens:
            sub_tokens = SplitTemplate(t, delim, delete_blanks)
            for st in sub_tokens:
                if hasattr(st, '__len__'):
                    if (len(st) > 0) or (not delete_blanks):
                        tokens_il.append(st)
                else:
                    tokens_il.append(st)
        tokens = tokens_il
        del tokens_il

    return tokens


def _TableFromTemplate(d, ltmpl, delimiters, delete_blanks):
    """
    See the docstring for the TableFromTemplate() function for an explanation.
    (This _TableFromTemplate() and SplitTemplate() are the workhorse functions
     for TableFromTemplate().)

    """

    output = SplitTemplateMulti(ltmpl, delimiters[d], delete_blanks[d])

    if d > 0:
        i = 0
        while i < len(output):
            output[i] = _TableFromTemplate(d - 1,
                                           output[i],
                                           delimiters,
                                           delete_blanks)
            # Delete empty LISTS?
            if (delete_blanks[d] and
                    hasattr(output[i], '__len__') and
                    (len(output[i]) == 0)):
                del output[i]
            else:
                i += 1

    return output


def TableFromTemplate(ltmpl, delimiters, delete_blanks=True):
    """
    This function can be used to split a template
    (a list containing TextBlocks and VarRefs) into a table
    into a multidimensional table, with an arbitrary number of dimensions.

    Arguments:

    ltmpl

    An alternating list of TextBlocks and VarRefs containing
    the contents of this text template.

    delimiters

    The user must supply a list or tuple of delimiters: one delimiter for
    each dimension in the table, with low-priority delimiters
    (such as spaces ' ') appearing first, and higher-priority delimiters
    (sich as newlines '\n') appearing later on in the list.
    This function will divide the entire "ltmpl" into an n-dimensional
    table.  Initially the text is split into a list of text using the
    highest-priority delimiter.  Then each entry in the resulting list is
    split into another list according to the next highest-priority delimiter.
    This continues until all of the delimiters are used up and an
    n-dimensional list-of-lists is remaining.

    delete_blanks

    The optional "delete_blanks" argument can be used to indicate whether
    or not to delete blank entries in the table (which occur as a result
    of placing two delimiters next to each other).  It should be either
    None (default), or it should be an array of booleans matching the
    size of the "delimiters" argument.  This allows the caller to customize
    the merge settings separately for each dimension (for example: to allow
    merging of whitespace within a line, without ignoring blank lines).


     ---- Details: ----

    1) Multi-character delimiters ARE allowed (like '\n\n').

    2) If a delimiter in the "delimiters" argument is not a string
    but is a tuple (or a list) of strings, then the text is split according
    to any of the delimiters in that tuple/list (starting from the last entry).
    This way, users can use this feature to split text according to multiple
    different kinds of whitespace characters (such as ' ' and '\t'), for
    example, buy setting delimiters[0] = (' ','\t').   If, additionally,
    delete_blanks[0] == True, then this will cause this function to
    divide text in without regard to whitespace on a given line (for example).

    Detailed example:

    table2D = TableFromTmplList(ltmpl,
                                delimiters = ((' ','\t'), '\n'),
                                delete_blanks = (True, False))

    This divides text in a similar way that the "awk" program does by default,
    ie, by ignoring various kinds of whitespace between text fields, but NOT
    ignoring blank lines.

    3) Any text contained in variable-names is ignored.

    """

    # Make a copy of ltmpl
    # (The workhorse function "_TableFromTemplate()" makes in-place changes to
    #  its "ltmpl" argument.  I don't want to modify "ltmpl", so I make a copy
    #  of it before I invoke "_TableFromTemplate()" on it.)

    output = [ltmpl[i] for i in range(0, len(ltmpl))]

    d = len(delimiters) - 1
    output = _TableFromTemplate(d, output, delimiters, delete_blanks)
    return output


class TemplateLexer(TtreeShlex):
    """ This class extends the standard python lexing module, shlex, adding a
    new member function (ReadTemplate()), which can read in a block of raw text,
    (halting at an (non-escaped) terminal character), and split the text into
    alternating blocks of text and variables.  (As far as this lexer is
    concerned, "variables" are simply tokens preceeded by $ or @ characters,
    and surrounded by optional curly-brackets {}.)

    """

    def __init__(self,
                 instream=None,
                 infile=None,
                 posix=False):
        TtreeShlex.__init__(self, instream, infile, posix)
        self.var_delim = '$@'  # characters which can begin a variable name
        self.var_open_paren = '{'  # optional parenthesis surround a variable
        self.var_close_paren = '}'  # optional parenthesis surround a variable
        self.newline = '\n'
        self.comment_skip_var = '#'

        #   Which characters belong in words?
        #
        # We want to allow these characters:
        #     ./$@&%^!*~`-_:;?<>[]()
        # to appear inside the tokens that TtreeShlex.get_token()
        # retrieves (TtreeShlex.get_token() is used to read class
        # names, and instance names, and variable names)
        #
        # settings.lex.wordchars+='./$@&%^!*~`-_+:;?<>[]' #Allow these chars
        #
        # Ommisions:
        # Note: I left out quotes, whitespace, comment chars ('#'), and escape
        #       characters ('\\') because they are also dealt with separately.
        #       Those characters should not overlap with settings.lex.wordchars.
        #
        # Enabling unicode support requires that we override this choice
        # by specifying "lex.wordterminators" instead of "wordchars".
        #
        # lex.wordterminators should be the (printable) set inverse of lex.wordchars
        # I'm not sure which ascii characters are NOT included in the string above
        # (We need to figure that out, and put them in settings.lex.wordterminators)
        # To figure that out, uncomment the 8 lines below:
        #
        # self.wordterminators=''
        # for i in range(0,256):
        #    c = chr(i)
        #    if c not in self.wordchars:
        #        self.wordterminators += c
        #sys.stderr.write('-------- wordterminators = --------\n')
        # sys.stderr.write(self.wordterminators+'\n')
        # sys.stderr.write('-----------------------------------\n')
        #
        # Here is the result:
        self.wordterminators = '(){|}' + \
            self.whitespace + \
            self.quotes + \
            self.operators + \
            self.escape + \
            self.commenters
        #  Note:
        # self.whitespace = ' \t\r\f\n'
        # self.quotes     = '\'"'
        # self.escape     = '\\'
        # self.commenters = '#'
        #  Note: I do not terminate on these characters: +-=*'"`
        # because they appear in the names of atom types in many force-fields.
        # Also * characters are needed for variables containing wildcards
        # in the name (which will be dealt with later).

        self.source_triggers = set(['include', 'import'])
        self.source_triggers_x = set(['import'])

    def GetSrcLoc(self):
        return OSrcLoc(self.infile, self.lineno)

    def ReadTemplate(self,
                     simplify_output=False,
                     terminators='}',
                     other_esc_chars='{',
                     keep_terminal_char=True):
        """
           ReadTemplate() reads a block of text (between terminators)
        and divides it into variables (tokens following a '$' or '@' character)
        and raw text.  This is similar to pythons string.Template(),
        however it reads from streams (files), not strings, and it allows use
        of more complicated variable names with multiple variable delimiters
        (eg '$' and '@').
        This readline()-like member function terminates when reaching a
        user-specified terminator character character (second argument),
        or when variable (eg: "$var"$ is encountered).  The result is
        a list of variable-separated text-blocks (stored in the first
        argument).   For example, the string:
        "string with $var1 and $var2 variables.}"  contains:
                "string with ",
                $var1,
                " and ",
                $var2,
                " variables.}"
        This simplifies the final process of rendering
        (substituting text into) the text blocks later on.
            Output:
        This function returns a list of (alternating) blocks of
        text, and variable names.  Each entry in the list is either:
        1) a text block:
               Raw text is copied from the source, verbatim, along with
               some additional data (filename and line numbers), to
               help retroactively identify where the text came from
               (in case a syntax error in the text is discovered later).
               In this case, the list entry is stored as a list
               The format (TextBlock) is similar to:
                  [text_string, ((filenameA,lineBegin), (filenameB,lineEnd))],
               where the tuples, (filenameA,lineBegin) and (filenameB,lineEnd)
               denote the source file(s) from which the text was read, and
               line number at the beginning and ending of the text block.
               (This information is useful for generating helpful error
               messages.  Note that the "TtreeShlex" class allows users to
               combine multiple files transparently into one stream using
               the "source" (or "sourcehook()") member.  For this reason, it
               is possible, although unlikely, that the text-block
               we are reading could span multiple different files.)
        2) a variable (for example "$var" or "${var}"):
               In this case, the list entry is stored in the "VarRef" format
               which is essentialy shown below:
                  [[var_prefix, var_nptr, var_suffix], (filename,lineno)]
               where var_prefix and var_suffix are strings containing brackets
               and other text enclosing the variable name (and may be empty).

       As an example, we consider a file named  "datafile" which
       contains the text containing 2 text blocks and 1 variable:
               "some\n text\n before ${var}. Text after\n".
       ReadTemplate() will read this and return a list with 3 entries:
             [ ['some\n text\n before', (('datafile', 1), ('datafile', 3))],
               [['${', 'var', '}'], ('datafile', 3, 3)],
               ['Text after\n', (('datafile', 3), ('datafile', 4))] ]

        Note that while parsing the text, self.lineno counter is
        incremented whenever a newline character is encountered.
        (Also: Unlike shlex.get_token(), this function does not
        delete commented text, or insert text from other files.)

            Exceptional Cases:
        Terminator characters are ignored if they are part of a variable
        reference. (For example, the '}' in "${var}", is used to denote a
        bracketed variable, and does not cause ReadTemplate() to stop reading)
           OR if they are part of a two-character escape sequence
        (for example, '}' in "\}" does not cause terminate parsing).
        In that case, the text is considered normal text.  (However the
        '\' character is also stripped out.  It is also stripped out if it
        preceeds any characters in "other_esc_chars", which is
        the second argument.  Otherwise it is left in the text block.)

        """
        #sys.stderr.write('    ReadTemplate('+terminators+') invoked at '+self.error_leader())

        # The main loop of the parser reads only one variable at time.
        # The following variables keep track of where we are in the template.
        reading_var = False  # Are we currently reading in the name of a variable?

        prev_char_delim = False  # True iff we just read a var_delim character like '$'
        # True iff we just read a (non-escaped) esc character '\'
        escaped_state = False
        # True iff we are in a region of text where vars should be ignored
        commented_state = False
        var_paren_depth = 0  # This is non-zero iff we are inside a
        # bracketed variable's name for example: "${var}"
        var_terminators = self.whitespace + self.newline + self.var_delim + '{}'

        tmpl_list = []  # List of alternating tuples of text_blocks and
        # variable names (see format comment above)
        # This list will be returned to the caller.

        # sys.stderr.write('report_progress='+str(report_progress))

        prev_filename = self.infile
        prev_lineno = self.lineno
        var_prefix = ''
        var_descr_plist = []
        var_suffix = ''
        text_block_plist = []

        done_reading = False

        while not done_reading:

            terminate_text = False
            terminate_var = False
            #delete_prior_escape = False

            nextchar = self.read_char()

            #sys.stderr.write('    ReadTemplate() nextchar=\''+nextchar+'\' at '+self.error_leader()+'  esc='+str(escaped_state)+', pvar='+str(prev_char_delim)+', paren='+str(var_paren_depth))

            # Count newlines:
            if nextchar in self.newline:
                commented_state = False
                self.lineno += 1

            elif ((nextchar in self.comment_skip_var) and
                  (not escaped_state)):
                commented_state = True

            # Check for end-of-file:
            if nextchar == '':

                if escaped_state:
                    raise InputError('Error: in ' + self.error_leader() + '\n\n'
                                     'No escaped character.')
                if reading_var:
                    terminate_var = True
                else:
                    terminate_text = True

                done_reading = True

            # --- Now process the character: ---

            # What we do next depends on which "mode" we are in.
            #  If we are reading a regular text block (reading_var == False),
            #   then we keep appending characters onto the end of "text_block",
            #   checking for terminal characters, or variable delimiters.
            #  If we are reading a variable name (reading_var == True),
            #   then we append characters to the end of "var_descr_plist[]",
            #   checking for variable terminator characters, as well as
            #   parenthesis (some variables are surrounded by parenthesis).

            elif reading_var:

                if nextchar in terminators:
                    #sys.stdout.write('   ReadTemplate() readmode found terminator.\n')
                    if escaped_state:
                        # In this case, the '\' char was only to prevent terminating
                        # string prematurely, so delete the '\' character.
                        #delete_prior_escape = True
                        if not (nextchar in self.var_close_paren):
                            del var_descr_plist[-1]
                            var_descr_plist.append(nextchar)

                    elif not ((var_paren_depth > 0) and (nextchar in self.var_close_paren)):
                        terminate_var = True
                        done_reading = True

                if nextchar in self.var_open_paren:  # eg: nextchar == '{'
                    #sys.stdout.write('   ReadTemplate() readmode found {.\n')
                    if escaped_state:
                        # In this case, the '\' char was only to prevent
                        # interpreting '{' as a variable prefix
                        # delete_prior_escape=True # so delete the '\'
                        # character
                        del var_descr_plist[-1]
                        var_descr_plist.append(nextchar)
                    else:
                        # "${var}" is a valid way to refer to a variable
                        if prev_char_delim:
                            var_prefix += nextchar
                            var_paren_depth = 1
                        # "${{var}}" is also a valid way to refer to a variable,
                        # (although strange), but "$va{r}" is not.
                        # Parenthesis (in bracketed variable names) must
                        # immediately follow the '$' character (as in "${var}")
                        elif var_paren_depth > 0:
                            var_paren_depth += 1

                elif nextchar in self.var_close_paren:
                    #sys.stdout.write('   ReadTemplate() readmode found }.\n')
                    if escaped_state:
                        # In this case, the '\' char was only to prevent
                        # interpreting '}' as a variable suffix,
                        # delete_prior_escape=True  #so skip the '\' character
                        if (nextchar not in terminators):
                            del var_descr_plist[-1]
                            var_descr_plist.append(nextchar)
                    else:
                        if var_paren_depth > 0:
                            var_paren_depth -= 1
                            if var_paren_depth == 0:
                                var_suffix = nextchar
                                terminate_var = True

                elif nextchar in var_terminators:
                    #sys.stdout.write('   ReadTemplate() readmode found var_terminator \"'+nextchar+'\"\n')
                    if (escaped_state or (var_paren_depth > 0)):
                        # In this case, the '\' char was only to prevent
                        # interpreting nextchar as a variable terminator
                        # delete_prior_escape = True # so skip the '\'
                        # character
                        del var_descr_plist[-1]
                        var_descr_plist.append(nextchar)
                    else:
                        terminate_var = True

                elif nextchar in self.var_delim:   # such as '$'
                    #sys.stdout.write('   ReadTemplate() readmode found var_delim.\n')
                    if escaped_state:
                        # In this case, the '\' char was only to prevent
                        # interpreting '$' as a new variable name
                        # delete_prior_escape = True # so skip the '\'
                        # character
                        del var_descr_plist[-1]
                        var_descr_plist.append(nextchar)
                    else:
                        prev_var_delim = True
                        # Then we are processing a new variable name
                        terminate_var = True
                else:
                    var_descr_plist.append(nextchar)
                    prev_char_delim = False

            else:  # begin else clause for "if reading_var:"

                # Then we are reading a text_block

                if nextchar in terminators:
                    if escaped_state:
                        # In this case, the '\' char was only to prevent terminating
                        # string prematurely, so delete the '\' character.
                        #delete_prior_escape = True
                        del text_block_plist[-1]
                        text_block_plist.append(nextchar)
                    elif commented_state:
                        text_block_plist.append(nextchar)
                    else:
                        terminate_text = True
                        done_reading = True

                elif nextchar in self.var_delim:   # such as '$'
                    if escaped_state:
                        # In this case, the '\' char was only to prevent
                        # interpreting '$' as a variable prefix.
                        # delete_prior_escape=True  #so delete the '\'
                        # character
                        del text_block_plist[-1]
                        text_block_plist.append(nextchar)
                    elif commented_state:
                        text_block_plist.append(nextchar)
                    else:
                        prev_char_delim = True
                        reading_var = True
                        var_paren_depth = 0
                        terminate_text = True
                else:
                    text_block_plist.append(nextchar)
                    # TO DO: use "list_of_chars.join()" instead of '+='
                    prev_char_delim = False  # the previous character was not '$'

            # Now deal with "other_esc_chars"
            # if escaped_state and (nextchar in other_esc_chars):

            if escaped_state and (nextchar in other_esc_chars):
                if reading_var:
                    #sys.stdout.write('   ReadTemplate: var_descr_str=\''+''.join(var_descr_plist)+'\'\n')
                    assert(var_descr_plist[-2] in self.escape)
                    del var_descr_plist[-2]
                else:
                    #sys.stdout.write('   ReadTemplate: text_block=\''+''.join(text_block_plist)+'\'\n')
                    assert(text_block_plist[-2] in self.escape)
                    del text_block_plist[-2]

            if terminate_text:
                #sys.stdout.write('ReadTemplate() appending: ')
                # sys.stdout.write(text_block)

                # tmpl_list.append( [text_block,
                #                   ((prev_filename, prev_lineno),
                #                    (self.infile, self.lineno))] )

                if simplify_output:
                    tmpl_list.append(''.join(text_block_plist))
                else:
                    tmpl_list.append(TextBlock(''.join(text_block_plist),
                                               OSrcLoc(prev_filename, prev_lineno)))
                    #, OSrcLoc(self.infile, self.lineno)))
                if not done_reading:
                    # The character that ended the text block
                    # was a variable delimiter (like '$'), in which case
                    # we should put it (nextchar) in the variable's prefix.
                    var_prefix = nextchar
                else:
                    var_prefix = ''
                var_descr_plist = []
                var_suffix = ''
                prev_filename = self.infile
                prev_lineno = self.lineno
                del text_block_plist
                text_block_plist = []
                # gc.collect()

            elif terminate_var:
                # Print an error if we terminated in the middle of
                # an incomplete variable name:
                if prev_char_delim:
                    raise InputError('Error: near ' + self.error_leader() + '\n\n'
                                     'Null variable name.')
                if var_paren_depth > 0:
                    raise InputError('Error: near ' + self.error_leader() + '\n\n'
                                     'Incomplete bracketed variable name.')

                var_descr_str = ''.join(var_descr_plist)

                # Now check for variable format modifiers,
                # like python's ".rjust()" and ".ljust()".
                # If present, then put these in the variable suffix.
                if ((len(var_descr_plist) > 0) and (var_descr_plist[-1] == ')')):
                    #i = len(var_descr_plist)-1
                    # while i >= 0:
                    #    if var_descr_plist[i] == '(':
                    #        break
                    #    i -= 1
                    i = var_descr_str.rfind('(')
                    if (((i - 6) >= 0) and
                        ((var_descr_str[i - 6:i] == '.rjust') or
                         (var_descr_str[i - 6:i] == '.ljust'))):
                        var_suffix = ''.join(
                            var_descr_plist[i - 6:]) + var_suffix
                        #var_descr_plist = var_descr_plist[:i-6]
                        var_descr_str = var_descr_str[:i - 6]

                # Process any special characters in the variable name
                var_descr_str = EscCharStrToChar(var_descr_str)

                # tmpl_list.append( [[var_prefix, var_descr_str, var_suffix],
                #                   (self.infile, self.lineno)] )
                if simplify_output:
                    tmpl_list.append(var_prefix + var_descr_str + var_suffix)
                else:
                    tmpl_list.append(VarRef(var_prefix, var_descr_str, var_suffix,
                                            OSrcLoc(self.infile, self.lineno)))

                # if report_progress:
                #sys.stderr.write('  parsed variable '+var_prefix+var_descr_str+var_suffix+'\n')

                #sys.stdout.write('ReadTemplate() appending: ')
                #sys.stderr.write(var_prefix + var_descr_str + var_suffix)

                del var_descr_plist
                del var_descr_str

                prev_filename = self.infile
                prev_lineno = self.lineno
                var_prefix = ''
                var_descr_plist = []
                var_suffix = ''
                # Special case: Variable delimiters like '$'
                #               terminate the reading of variables,
                #               but they also signify that a new
                #               variable is being read.
                if nextchar in self.var_delim:
                    # Then we are processing a new variable name
                    prev_var_delim = True
                    reading_var = True
                    var_paren_depth = 0
                    var_prefix = nextchar

                elif nextchar in self.var_close_paren:
                    del text_block_plist
                    text_block_plist = []
                    # gc.collect()
                    prev_var_delim = False
                    reading_var = False

                else:
                    # Generally, we don't want to initialize the next text block
                    # with the empty string.  Consider that whatever character
                    # caused us to stop reading the previous variable and append
                    # it to the block of text that comes after.
                    del text_block_plist
                    text_block_plist = [nextchar]
                    # gc.collect()
                    prev_var_delim = False
                    reading_var = False

            # If we reached the end of the template (and the user requests it),
            # then the terminal character can be included in the list
            # of text_blocks to be returned to the caller.
            if done_reading and keep_terminal_char:
                #sys.stdout.write('ReadTemplate() appending: \''+nextchar+'\'\n')
                # Here we create a new text block which contains only the
                # terminal character (nextchar).
                # tmpl_list.append( [nextchar,
                #                   ((self.infile, self.lineno),
                #                    (self.infile, self.lineno))] )
                if simplify_output:
                    tmpl_list.append(nextchar)
                else:
                    tmpl_list.append(TextBlock(nextchar,
                                               OSrcLoc(self.infile, self.lineno)))
                    #, OSrcLoc(self.infile, self.lineno)))

            if escaped_state:
                escaped_state = False
            else:
                if nextchar in self.escape:
                    escaped_state = True

        #sys.stderr.write("*** TMPL_LIST0  = ***", tmpl_list)
        return tmpl_list  # <- return value stored here

    def GetParenExpr(self, prepend_str='', left_paren='(', right_paren=')'):
        """ GetParenExpr() is useful for reading in strings
            with nested parenthesis and spaces.
            This function can read in the entire string:

              .trans(0, 10.0*sin(30), 10.0*cos(30))

            (Because I was too lazy to write this correctly...)
            Spaces are currently stripped out of the expression.
            (...unless surrounded by quotes) The string above becomes:

              ".trans(0,10.0*sin(30),10.0*cos(30))"

            Sometimes the caller wants to prepend some text to the beginning
            of the expression (which may contain parenthesis).  For this
            reason, an optional first argument ("prepend_str") can be
            provided.  By default it is empty.

        """

        src_loc_begin = SrcLoc(self.infile, self.lineno)
        orig_wordterm = self.wordterminators
        self.wordterminators = self.wordterminators.replace(
            left_paren, '').replace(right_paren, '')

        token = self.get_token()
        if ((token == '') or
                (token == self.eof)):
            return prepend_str

        expr_str = prepend_str + token

        # if (expr_str.find(left_paren) == -1):
        #    raise InputError('Error near or before '+self.error_leader()+'\n'
        #                     'Expected an open-paren (\"'+prepend_str+left_paren+'\") before this point.\n')
        #    return expr_str

        paren_depth = expr_str.count(left_paren) - expr_str.count(right_paren)
        while ((len(expr_str) == 0) or (paren_depth > 0)):
            token = self.get_token()
            if ((type(token) is not str) or
                    (token == '')):
                raise InputError('Error somewhere between ' +
                                 self.error_leader(src_loc_begin.infile,
                                                   src_loc_begin.lineno)
                                 + 'and ' + self.error_leader() + '\n'
                                 'Invalid expression: \"' + expr_str[0:760] + '\"')
            expr_str += token
            paren_depth = expr_str.count(
                left_paren) - expr_str.count(right_paren)
        if (paren_depth != 0):
            raise InputError('Error somewhere between ' +
                             self.error_leader(src_loc_begin.infile,
                                               src_loc_begin.lineno)
                             + 'and ' + self.error_leader() + '\n'
                             'Invalid expression: \"' + expr_str[0:760] + '\"')
        self.wordterminators = orig_wordterm
        return expr_str


if __name__ == '__main__':
    if len(sys.argv) == 1:
        lexer = TtreeShlex()
    else:
        file = sys.argv[1]
        lexer = TtreeShlex(open(file), file)
    while 1:
        tt = lexer.get_token()
        if tt:
            sys.stderr.write("Token: " + repr(tt))
        else:
            break



#!/usr/bin/env python

man_page_text = """
Usage (example):

postprocess_coeffs.py ttree_assignments.txt < file.template > file_new.template

Moltemplate users would like to be able to use commands with wildcards like:
  bond_coeff @bond:*/A/* harmonic 533.7 120.0
  pair_coeff @atom:A* @atom:B* lj/cut 0.35 2.7
as shorthand for specifying force field parameters for multiple atom&bond types:
  bond_coeff @bond:a/A/c harmonic 533.7 120.0
  bond_coeff @bond:b/A/d harmonic 533.7 120.0
  pair_coeff @atom:A* @atom:B* lj/cut 0.35 2.7
  pair_coeff @atom:A  @atom:B  lj/cut 0.35 2.7
  pair_coeff @atom:A1 @atom:B1 lj/cut 0.35 2.7
  pair_coeff @atom:A1 @atom:B2 lj/cut 0.35 2.7
  pair_coeff @atom:A2 @atom:B1 lj/cut 0.35 2.7
  pair_coeff @atom:A2 @atom:B2 lj/cut 0.35 2.7
      :         :        :       :     :    :
However LAMMPS does not interpret the * character this way.
Hence, this script will replace the line with the * wildcards above with the 
lines text that follow above.  This script also works for bond_coeff, 
angle_coeff, dihedral_coeff, and improper_coeff commands (expanding wildcard
characters appearing in @bond, @angle, @dihedral, and @improper variables).

This program expects an argument (ttree_assignments.txt) which has a list of all
of the atom types (and bond_types, and angle_types...) which have been defined
and it will substitute those variable names in place of the wildcard expressions
(The ttree_assignments.txt file is a 2-column file variables in the 1st column, 
 and their integer values in the 2nd column.  The 2nd column is ignored.)

The resulting list of commands with explicit variable names will be 
printed to the standard-out.

"""


import sys
import gc

try:
    from .ttree import ExtractFormattingCommands
    from .ttree_lex import *

except (SystemError, ValueError):
    # not installed as a package
    from ttree import ExtractFormattingCommands
    from ttree_lex import *


g_filename = __file__.split('/')[-1]
g_module_name = g_filename
if g_filename.rfind('.py') != -1:
    g_module_name = g_filename[:g_filename.rfind('.py')]
g_date_str = '2017-8-31'
g_version_str = '0.1.0'
g_program_name = g_filename
#sys.stderr.write(g_program_name+' v'+g_version_str+' '+g_date_str+' ')


def ExtractVarName(text):
    """ Read a string like ' @atom:A  '  or  '  @{/atom:A B/C/../D }ABC '
        and return (' ','@atom:A','  ')  or  ('  ','@{/atom:A B/C/../D }','ABC')
        These are 3-tuples containing the portion of the text containing 
        only the variable's name (assumed to be within the text),
        ...in addition to the text on either side of the variable name.
    """
    i_begin = 0
    escape = '\''
    lparen = '{'
    rparen = '}'
    escaped = False
    commenters = '#'
    whitespace = ' \t\r\f\n'
    terminators = whitespace + commenters
    # Ideally, perhaps I should lookup these values from ttree_lex.TtreeLex to 
    # make sure I am being consistent, instead of re-defining them in this file.
    in_paren = text[0:1] == lparen
    if in_paren:
        terminators = rparen
    i_begin = 0
    while ((i_begin < len(text)) and
           (text[i_begin] in whitespace)):
        i_begin += 1
    i_end = i_begin
    while ((i_end < len(text)) and
           (text[i_end] not in terminators)):
        i_end += 1
    return (text[0: i_begin],
            text[i_begin:i_end],
            text[i_end:])



def main():
    try:
        if (len(sys.argv) != 2):
            raise InputError('Error running  \"' + g_program_name + '\"\n'
                             ' Typical usage:\n'
                             ' postprocess_coeffs.py ttree_assignments.txt < file.template > file.rendered\n'
                             '\n'
                             '   Missing argument.\n'
                             '   Expected the name of a 2-column file containing\n'
                             '   variable names and their bindings (values).\n'
                             '   (This is likely a programmer error.\n'
                             '    This script was not intended to be run by end users.)\n')

        bindings_filename = sys.argv[1]
        f = open(bindings_filename)
        atom_types = set([])
        bond_types = set([])
        angle_types = set([])
        dihedral_types = set([])
        improper_types = set([])

        #BasicUIReadBindingsStream(assignments, f, bindings_filename)

        # The line above is robust but it uses far too much memory.
        # This for loop below works for most cases.
        for line in f:
            #tokens = lines.strip().split()
            # like split but handles quotes
            tokens = SplitQuotedString(line.strip())
            if len(tokens) < 2:
                continue
            if tokens[0].find('@') != 0:
                continue
            if tokens[0][2:].find('atom') == 0:
                atom_types.add(tokens[0][1:])
            elif tokens[0][2:].find('bond') == 0:
                bond_types.add(tokens[0][1:])
            elif tokens[0][2:].find('angle') == 0:
                angle_types.add(tokens[0][1:])
            elif tokens[0][2:].find('dihedral') == 0:
                dihedral_types.add(tokens[0][1:])
            elif tokens[0][2:].find('improper') == 0:
                improper_types.add(tokens[0][1:])

        f.close()
        gc.collect()

        lex = LineLex(sys.stdin, '__standard_input_for_postprocess_coeffs__')
        #lex = LineLex(open('deleteme.template', 'r'), '__standard_input_for_postprocess_coeffs_')
        lex.commenters = ''            #(don't attempt to skip over comments)
        lex.line_extend_chars += '&'   #(because LAMMPS interprets '&' as '\')

        while True:
            line_orig = lex.ReadLine()
            #sys.stderr.write('line_orig = \"'+str(line_orig)+'\"\n')
            if (not line_orig) or (line_orig == ''):
                break
            tokens = line_orig.strip().split('@')
            if ((len(tokens) >= 2) and
                (tokens[0].find('bond_coeff') == 0) and
                HasWildcard(tokens[1])): #does this token contain '*' or '?'
                text_before, typepattern, text_after = ExtractVarName(tokens[1])
                for btype in bond_types:
                    if MatchesPattern(btype, typepattern):
                        assert(text_before == '')
                        tokens[1] = btype + text_after
                        sys.stdout.write('@'.join(tokens) + '\n')
            elif ((len(tokens) >= 2) and
                  (tokens[0].find('angle_coeff') == 0) and
                  HasWildcard(tokens[1])): #does this token contain '*' or '?'
                text_before, typepattern, text_after = ExtractVarName(tokens[1])
                for antype in angle_types:
                    if MatchesPattern(antype, typepattern):
                        assert(text_before == '')
                        tokens[1] = antype + text_after
                        sys.stdout.write('@'.join(tokens) + '\n')
            elif ((len(tokens) >= 2) and
                  (tokens[0].find('dihedral_coeff') == 0) and
                  HasWildcard(tokens[1])): #does this token contain '*' or '?'
                text_before, typepattern, text_after = ExtractVarName(tokens[1])
                for dtype in dihedral_types:
                    if MatchesPattern(dtype, typepattern):
                        assert(text_before == '')
                        tokens[1] = dtype + text_after
                        sys.stdout.write('@'.join(tokens) + '\n')
            elif ((len(tokens) >= 2) and
                  (tokens[0].find('improper_coeff') == 0) and
                  HasWildcard(tokens[1])): #does this token contain '*' or '?'
                text_before, typepattern, text_after = ExtractVarName(tokens[1])
                for itype in improper_types:
                    if MatchesPattern(itype, typepattern):
                        assert(text_before == '')
                        tokens[1] = itype + text_after
                        sys.stdout.write('@'.join(tokens) + '\n')
            #elif ((len(tokens) >= 3) and
            #      (tokens[0].find('pair_coeff') == 0) and
            #      (HasWildcard(tokens[1]) or HasWildcard(tokens[2]))):
            elif ((len(tokens) >= 2) and
                  (tokens[0].find('pair_coeff') == 0)):
                # First deal with cases with only one @variable, such as:
                #    pair_coeff @atom:A*   *       ...
                #    pair_coeff    *     @atom:A*  ...
                # (We don't deal with cases like "* *" because LAMMPS interprets
                #  these in a special way:  manybody pair_styles use "* *")
                if len(tokens) == 2:
                    if tokens[0].rstrip()[-1:] == '*':
                        tokens[0] = tokens[0].rstrip()[:-1]
                        tokens.insert(1, '/atom:* ')
                    else:
                        ic = tokens[1].find(' * ')
                        tokens.append('/atom:* '+tokens[1][ic+3:])
                        tokens[1] = tokens[1][:ic]+' '
                assert(len(tokens) >= 3)
                text_before1,typepattern1,text_after1=ExtractVarName(tokens[1])
                text_before2,typepattern2,text_after2=ExtractVarName(tokens[2])
                if HasWildcard(tokens[1]):
                    atom_types1 = atom_types
                else:
                    atom_types1 = set([typepattern1])
                if HasWildcard(tokens[2]):
                    atom_types2 = atom_types
                else:
                    atom_types2 = set([typepattern2])
                for atype1 in atom_types1:
                    #sys.stderr.write('atype1 = \"'+str(atype1)+'\"\n')
                    if MatchesPattern(atype1, typepattern1):
                        assert(text_before1 == '')
                        tokens[1] = atype1 + text_after1
                        for atype2 in atom_types2:
                            #sys.stderr.write(' atype2 = \"'+str(atype2)+'\"\n')
                            if MatchesPattern(atype2, typepattern2):
                                assert(text_before2 == '')
                                tokens[2] = atype2 + text_after2
                                sys.stdout.write('@'.join(tokens) + '\n')
            else:
                sys.stdout.write(line_orig)

    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

    return

if __name__ == '__main__':
    main()



try:
    from ..nbody_graph_search import Ugraph
except:
    # not installed as a module
    from nbody_graph_search import Ugraph

# This file defines how improper interactions are generated in AMBER (GAFF).
# To use it, add "(gaff_imp.py)" to the name of the "Data Impropers By Type"
# section, and make sure this file is located in the "common" directory.
# For example:
# write_once("Data Impropers By Type (gaff_imp.py)") {
#   ...
# }


#    To find 4-body "improper" interactions,
#    (by default, most of the time), we would use this subgraph:
#           0
#           *                  1st bond connects atoms 2 and 0
#           |              =>  2nd bond connects atoms 2 and 1
#         _.*._                3rd bond connects atoms 2 and 3
#       *'  2  `*
#      1         3
#
# In AMBER/GAFF, the central atom is the third atom ("2").
# http://archive.ambermd.org/201307/0519.html
# This differs from other force-fields.
# We take this detail into account in the line below:

bond_pattern = Ugraph([(2,0), (2,1), (2,3)])

# As with other force-fields, the improper-angle is the angle between the planes
# defined by the first three atoms (0,1,2) and last three atoms (1,2,3).
# (This is implemented in LAMMPS using an improper_style which requires
#  that the atoms in the interaction will be listed in this order: 0,1,2,3.)

def canonical_order(match):
    """
    Before defining a new interaction, we must check to see if an
    interaction between these same 4 atoms has already been created
     (perhaps listed in a different, but equivalent order).
    If we don't check for this this, we will create many unnecessary redundant
    interactions (which can slow down he simulation).
    To avoid this, I define a "canonical_order" function which sorts the atoms
    and bonds in a way which is consistent with the symmetry of the interaction
    being generated...  Later the re-ordered list of atom and bond ids will be
    tested against the list of atom/bond ids in the matches-found-so-far,
    before it is added to the list of interactions found so far.  Note that
    the energy of an improper interactions is a function of the improper angle.
    The "improper angle" is often defined as the angle between planes formed
    by atoms 0,1,2 & 1,2,3.  (Alternately, it is sometimes defined as the
    angle between the 0,1,2 plane and atom 3.)
    This angle does not change when swapping the OUTER pair of atoms (0 and 3)
    (except for a change of sign, which does not matter since the energy functions
     used are typically sign invariant.  Furthermore, neither of OUTER pair of atoms
     are the central atom. There are 3!=6 ways of ordering the remaining 3 atoms.)
    Consequently it does not make sense to define a separate 4-body improper-
    interaction between atoms 0,1,2,3   AS WELL AS between  3,1,2,0.
    So we sort the atoms and bonds so that the first atom has a always has
    a lower atomID than the last atom.  (Later we will check to see if we
    have already defined an interaction between these 4 atoms.  If not then
    we create a new one.)

    """
    atom0 = match[0][0]
    atom1 = match[0][1]
    atom2 = match[0][2]
    atom3 = match[0][3]
    # match[1][0:2] contains the ID numbers for the 3 bonds
    bond0 = match[1][0]
    bond1 = match[1][1]
    bond2 = match[1][2]
    if atom0 <= atom3:
        #return ((atom0,atom1,atom2,atom3), (bond0, bond1, bond2))
        # But this is the same thing as:
        return match
    else:
        return ((atom3,atom1,atom2,atom0), (bond2,bond1,bond0))



#!/usr/bin/env python

"""
nbody_fix_ttree_assignments.py

This is an ugly little script which was not intended to be run by end users.

Typical usage:

nbody_fix_ttree_assignments.py "angles" new_Angles.template \
  < ttree_assignments.txt > ttree_assigmnents_new.txt

What it does:

In this example, this program extracts the first column from
"new_Angles.template", and appends to it the first column from
the lines in ttree_assignments.txt containing "@angle:".
Then it adds a second column which is just a sequence of integers
counting upwards.
Finally it inserts this 2-column text into the appropriate place in a
ttree_assignments.txt file, replacing the original @angle variables
with the new ones (followed by the renumbered original).


   AWK/GREP equivalent
This program is roughly equivalent to the following lines of awk/grep:

awk 'BEGIN{i=-1} {if(substr($0,0,8)=="$/angle") i=NR; if (i==-1){print $0}}'\
    < ttree_assignments.txt > ttree_assignments_new.txt

awk '{print $1}' < new_Angles.template > Angles_column1.txt
grep '$/angle:' ttree_assignments.txt | awk '{print $1}' >> Angles_column1.txt
awk '{print $1 "  " NR}' <  Angles_column1.txt >> ttree_assignments_new.txt

awk 'BEGIN{found=0;passed=0} {if(substr($0,0,8)=="$/angle") found=1;
                              else {if (found) {passed=1}}
                              if (passed) print $0}' \
          < ttree_assignments.txt >> ttree_assignments_new.txt

I wrote this python script (instead of using awk) just to handle quoted stings
(and strings with other fancy characters and escape sequences).

"""

import sys

try:
    from .ttree_lex import SplitQuotedString, InputError
except (SystemError, ValueError):
    # not installed as a package
    from ttree_lex import *

g_program_name = __file__.split('/')[-1]


def main():
    try:
        if (len(sys.argv) != 3):
            raise InputError('Error running  \"' + g_program_name + '\"\n'
                             '   Wrong number of arguments.\n'
                             '   (This is likely a programmer error.\n'
                             '    This script was not intended to be run by end users.)\n')

        cat_name = sys.argv[1]
        f = open(sys.argv[2])
        lines_generated = f.readlines()
        f.close()

        # Selections are simply lists of 2-tuples (pairs)
        #f = open('ttree_assignments.txt','r')
        #lines_bindings = f.readlines()
        # f.close()
        lines_bindings = sys.stdin.readlines()

        # Figure out which lines in the 'ttree_assignments.txt' file
        # contain the variables of the type you are looking for.
        # Make note of the relevant line numbers
        i_preexisting_begin = -1
        i_preexisting_end = -1
        in_section = False
        possible_cat_names = set(
            ['$' + cat_name, '$/' + cat_name, '${' + cat_name, '${/' + cat_name])

        preexisting_interaction_list = []
        for i in range(0, len(lines_bindings)):
            line = lines_bindings[i].strip()
            tokens = SplitQuotedString(line)  # strip comments, handle quotes
            if len(tokens) == 2:
                before_colon = tokens[0].split(':')[0]
                if before_colon in possible_cat_names:
                    if i_preexisting_begin == -1:
                        i_preexisting_begin = i
                        in_section = True
                else:
                    if in_section:
                        i_preexisting_end = i
                    in_section = False

        if i_preexisting_end == -1:
            i_preexisting_end = len(lines_bindings)

        if i_preexisting_begin == -1:
            for line in lines_bindings:
                sys.stdout.write(line)
        else:
            # write out all the lines in the original file up until the point where
            # the variables in the category we are looking for were encountered
            for i in range(0, i_preexisting_begin):
                sys.stdout.write(lines_bindings[i])

        sys.stderr.write('  (adding new lines)\n')

        # Now add some new lines (2-column format).
        # As with any ttree_assignment.txt file:
        #   The first column has our generated variable names
        #   The second column has the counter assigned to that variable
        new_counter = 1
        for line_orig in lines_generated:
            line = line_orig.strip()
            if len(line) > 0:
                tokens = SplitQuotedString(line)  # strip comments, handle quotes
                sys.stdout.write(tokens[0] + '  ' + str(new_counter) + '\n')
                new_counter += 1

        sys.stderr.write('  (adding pre-exisiting lines)\n')
        if i_preexisting_begin != -1:
            # Append the original pre-existing interactions of that type, but assign
            # them to higher numbers.  (Hopefully this helps to make sure that these
            # assignments will override any of the automatic/generated assignments.)
            # As with any ttree_assignment.txt file:
            #   The first column has our generated variable names
            #   The second column has the counter assigned to that variable

            # sys.stderr.write('  i_preexisting_begin='+
            #                 str(i_preexisting_begin)+
            #                 ' i_preexisting_end='+str(i_preexisting_end)+'\n')

            for i in range(i_preexisting_begin, i_preexisting_end):
                line = lines_bindings[i].strip()
                tokens = SplitQuotedString(line)  # strip comments, handle quotes
                if len(tokens) == 2:
                    sys.stdout.write(tokens[0] + '  ' + str(new_counter) + '\n')
                    new_counter += 1

            #sys.stderr.write('  (writing pre-exisiting lines)\n')

            # write out all the lines in the original file after this point.
            for i in range(i_preexisting_end, len(lines_bindings)):
                sys.stdout.write(lines_bindings[i])

        sys.exit(0)


    except (ValueError, InputError) as err:
        sys.stderr.write('\n' + str(err) + '\n')
        sys.exit(-1)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import sys

def main():
    for line_orig in sys.stdin:
        line = line_orig.rstrip('\n')
        comment = ''
        if '#' in line_orig:
            ic = line.find('#')
            line = line_orig[:ic]
            comment = ' '+line_orig[ic:].rstrip('\n')

        tokens = line.strip().split()
        if len(tokens) > 2:
            atomid = -1
            atomtype = -1
            pos_found = False
            for i in range(0,len(tokens)):
                if (tokens[i] == 'part') and (i+1 < len(tokens)):
                    atomid = tokens[i+1]
                elif (tokens[i] == 'type') and (i+1 < len(tokens)):
                    atomtype = tokens[i+1]
                elif (tokens[i] == 'pos') and (i+2 < len(tokens)):
                    pos_found = True
            if (atomid != -1) and (atomtype != -1) and pos_found:
                sys.stdout.write(atomid+' '+atomtype+'\n')

if __name__ == "__main__":
    main()



from .ttree import BasicUISettings, BasicUIParseArgs, EraseTemplateFiles, \
    StackableCommand, PopCommand, PopRightCommand, PopLeftCommand, \
    PushCommand, PushLeftCommand, PushRightCommand, ScopeCommand, \
    WriteVarBindingsFile, StaticObj, InstanceObj, \
    BasicUI, ScopeBegin, ScopeEnd, WriteFileCommand, Render
from .ttree_lex import TtreeShlex, split, LineLex, SplitQuotedString, \
    EscCharStrToChar, SafelyEncodeString, RemoveOuterQuotes, MaxLenStr, \
    HasWildcard, InputError, ErrorLeader, SrcLoc, OSrcLoc, TextBlock, VarRef, \
    VarNPtr, VarBinding, SplitTemplate, SplitTemplateMulti, TableFromTemplate, \
    ExtractCatName, DeleteLinesWithBadVars, TemplateLexer
from .lttree_styles import AtomStyle2ColNames, ColNames2AidAtypeMolid, \
    ColNames2Coords, ColNames2Vects, ColNames2Vects, data_atoms, data_masses
from .ettree_styles import \
    LinesWSlashes, SplitMultiDelims, SplitAtomLine, \
    iEsptAtomCoords, iEsptAtomVects, iEsptAtomType, iEsptAtomID
from .ttree_matrix_stack import MultMat, MatToStr, LinTransform, \
    AffineTransform, AffineCompose, CopyMat, ScaleMat, RotMatAXYZ, \
    CrossProd, DotProd, Length, Normalize, RotMatXYZXYZ, MultiAffineStack
